import pandas as pd

from meteostat.api.point import Point
from meteostat.api.timeseries import TimeSeries


def nearest_neighbor(df: pd.DataFrame, ts: TimeSeries, _point: Point) -> pd.DataFrame:
    """
    Get nearest neighbor value for each record in a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing the data to be adjusted.
    ts : TimeSeries
        TimeSeries object containing the target data.
    _point : Point
        Point object representing the target location.

    Returns
    -------
    pd.DataFrame
        DataFrame with nearest neighbor values for each record.
    """
    df = (
        df.sort_values("distance")
        .groupby(pd.Grouper(level="time", freq=ts.freq))
        .agg("first")
    )

    return df



"""
Monthly Time Series Data

Access monthly time series data for one or multiple weather stations.
"""

from typing import List, Optional
from datetime import datetime, date

import pandas as pd

from meteostat.core.data import data_service
from meteostat.enumerations import Parameter, Provider, Granularity
from meteostat.typing import Station, Request
from meteostat.api.point import Point
from meteostat.utils.parsers import parse_station, parse_time

DEFAULT_PARAMETERS = [
    Parameter.TEMP,
    Parameter.TMIN,
    Parameter.TMAX,
    Parameter.TXMN,
    Parameter.TXMX,
    Parameter.PRCP,
    Parameter.PRES,
    Parameter.TSUN,
]


def monthly(
    station: str | Station | Point | List[str | Station | Point] | pd.DataFrame,
    start: Optional[datetime | date],
    end: Optional[datetime | date],
    parameters: Optional[List[Parameter]] = None,
    providers: Optional[List[Provider]] = None,
):
    """
    Access monthly time series data.

    Parameters
    ----------
    station : str, Station, Point, List[str | Station | Point], pd.Index, pd.Series
        Weather station(s) or Point(s) to query data for. Can be a single station/point or a list.
        Points are converted to virtual stations with IDs like $0001, $0002, etc.
    start : datetime, date, optional
        Start date for the data query. If None, the earliest available date will be used.
    end : datetime, date, optional
        End date for the data query. If None, the latest available date will be used.
    parameters : List[Parameter], optional
        List of parameters to include in the data query. Defaults to a set of common parameters.
    providers : List[Provider], optional
        List of data providers to use for the query. Defaults to the monthly provider.

    Returns
    -------
    TimeSeries
        A TimeSeries object containing the monthly data for the specified stations and parameters.
    """
    if parameters is None:
        parameters = DEFAULT_PARAMETERS
    if providers is None:
        providers = [Provider.MONTHLY]

    req = Request(
        granularity=Granularity.MONTHLY,
        providers=providers,
        parameters=parameters,
        station=parse_station(station),
        start=parse_time(start),
        end=parse_time(end, is_end=True),
    )

    return data_service.fetch(req)



from datetime import datetime
from typing import Optional

import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.core.cache import cache_service
from meteostat.core.network import network_service
from meteostat.providers.eccc.shared import ENDPOINT, get_meta_data
from meteostat.typing import ProviderRequest

BATCH_LIMIT = 9000
PROPERTIES = {
    "LOCAL_DATE": "time",
    "MAX_TEMPERATURE": Parameter.TMAX,
    "MEAN_TEMPERATURE": Parameter.TEMP,
    "MIN_TEMPERATURE": Parameter.TMIN,
    "SPEED_MAX_GUST": Parameter.WPGT,
    "TOTAL_PRECIPITATION": Parameter.PRCP,
    "SNOW_ON_GROUND": Parameter.SNWD,
    "TOTAL_SNOW": Parameter.SNOW,
}


@cache_service.cache(TTL.DAY, "pickle")
def get_df(climate_id: str, year: int) -> Optional[pd.DataFrame]:
    # Process start & end date
    # ECCC uses the station's local time zone
    start = datetime(year, 1, 1, 0, 0, 0).strftime("%Y-%m-%dT%H:%M:%S")
    end = datetime(year, 12, 31, 23, 59, 59).strftime("%Y-%m-%dT%H:%M:%S")

    response = network_service.get(
        f"{ENDPOINT}/collections/climate-daily/items",
        params={
            "CLIMATE_IDENTIFIER": climate_id,
            "datetime": f"{start}/{end}",
            "f": "json",
            "properties": ",".join(PROPERTIES.keys()),
            "limit": BATCH_LIMIT,
        },
    )

    data = response.json()

    # Extract features from the response
    features = map(
        lambda feature: feature["properties"] if "properties" in feature else {},
        data.get("features", []),
    )

    # Create a DataFrame from the extracted features
    df = pd.DataFrame(features)
    df = df.rename(columns=PROPERTIES)

    # Handle time column & set index
    df["time"] = pd.to_datetime(df["time"])
    df = df.set_index(["time"])

    return df


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if (
        "national" not in req.station.identifiers
        or req.start is None
        or req.end is None
    ):
        return None

    meta_data = get_meta_data(req.station.identifiers["national"])
    climate_id = meta_data.get("CLIMATE_IDENTIFIER")
    archive_first = meta_data.get("DLY_FIRST_DATE")
    archive_last = meta_data.get("DLY_LAST_DATE")

    if not (climate_id and archive_first and archive_last):
        return None

    archive_start = datetime.strptime(archive_first, "%Y-%m-%d %H:%M:%S")
    archive_end = datetime.strptime(archive_last, "%Y-%m-%d %H:%M:%S")

    years = range(
        max(req.start.year, archive_start.year),
        min(req.end.year, archive_end.year) + 1,
    )
    data = [get_df(climate_id, year) for year in years]

    return pd.concat(data) if len(data) and not all(d is None for d in data) else None



"""
Validators Module

Provides validator functions for data validation.
"""

from pandas import Series

from meteostat.core.validator import Validator


def minimum(value: int | float) -> Validator:
    """
    Numeric minimum
    """

    def _func(series: Series) -> Series:
        return series >= value

    return Validator(_func)


def maximum(value: int | float) -> Validator:
    """
    Numeric maximum
    """

    def _func(series: Series) -> Series:
        return series <= value

    return Validator(_func)



from typing import Optional

import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.core.cache import cache_service
from meteostat.core.network import network_service
from meteostat.providers.eccc.shared import ENDPOINT, get_meta_data
from meteostat.typing import ProviderRequest

BATCH_LIMIT = 9000
PROPERTIES = {
    "LOCAL_DATE": "time",
    "MAX_TEMPERATURE": Parameter.TXMX,
    "MEAN_TEMPERATURE": Parameter.TEMP,
    "MIN_TEMPERATURE": Parameter.TXMN,
    "TOTAL_PRECIPITATION": Parameter.PRCP,
    "DAYS_WITH_PRECIP_GE_1MM": Parameter.PDAY,
    "TOTAL_SNOWFALL": Parameter.SNOW,
}


@cache_service.cache(TTL.WEEK, "pickle")
def get_df(
    climate_id: str,
) -> Optional[pd.DataFrame]:
    response = network_service.get(
        f"{ENDPOINT}/collections/climate-monthly/items",
        params={
            "CLIMATE_IDENTIFIER": climate_id,
            "f": "json",
            "properties": ",".join(PROPERTIES.keys()),
            "limit": BATCH_LIMIT,
        },
    )

    data = response.json()

    # Extract features from the response
    features = map(
        lambda feature: feature["properties"] if "properties" in feature else {},
        data.get("features", []),
    )

    # Create a DataFrame from the extracted features
    df = pd.DataFrame(features)
    df = df.rename(columns=PROPERTIES)

    # Handle time column & set index
    df["time"] = pd.to_datetime(df["time"])
    df = df.set_index(["time"])

    return df


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if "national" not in req.station.identifiers:
        return None

    meta_data = get_meta_data(req.station.identifiers["national"])
    climate_id = meta_data.get("CLIMATE_IDENTIFIER")

    if not climate_id:
        return None

    return get_df(climate_id)



"""
DWD MOSMIX data provider

Parameters: https://www.dwd.de/DE/leistungen/met_verfahren_mosmix/mosmix_parameteruebersicht.pdf?__blob=publicationFile&v=3
"""

import re
from io import BytesIO
from typing import Optional
from datetime import datetime
from zipfile import ZipFile
from lxml import etree  # type: ignore

import pandas as pd

from meteostat.core.cache import cache_service
from meteostat.enumerations import TTL, Parameter
from meteostat.typing import ProviderRequest
from meteostat.utils.conversions import (
    kelvin_to_celsius,
    ms_to_kmh,
    percentage_to_okta,
    temp_dwpt_to_rhum,
)
from meteostat.core.network import network_service

ENDPOINT = "https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_L/single_stations/{station}/kml/MOSMIX_L_LATEST_{station}.kmz"
COCO_MAP = {
    "0": 1,
    "1": 2,
    "2": 3,
    "3": 4,
    "45": 5,
    "49": 5,
    "61": 7,
    "63": 8,
    "65": 9,
    "51": 7,
    "53": 8,
    "55": 9,
    "68": 12,
    "69": 13,
    "71": 14,
    "73": 15,
    "75": 16,
    "80": 17,
    "81": 18,
    "82": 18,
    "83": 19,
    "84": 20,
    "85": 21,
    "86": 22,
    "66": 10,
    "67": 11,
    "56": 10,
    "57": 11,
    "95": 25,
}


def get_coco(code: str | int) -> Optional[int]:
    """
    Map DWD MOSMIX weather condition codes to Meteostat condicodes
    """
    return COCO_MAP.get(str(code))


@cache_service.cache(TTL.HOUR, "pickle")
def get_df(station: str) -> Optional[pd.DataFrame]:
    # Fetch the KMZ file data in memory
    response = network_service.get(ENDPOINT.format(station=station))
    kmz_data = BytesIO(response.content)

    # KMZ -> KML in memory
    with ZipFile(kmz_data, "r") as kmz:
        with kmz.open(kmz.infolist()[0].filename, "r") as raw:
            kml = raw.read()

    # Parse KML
    tree = etree.fromstring(kml)

    # Skip stale forecasts
    issue_time = datetime.strptime(
        tree.xpath(
            "//kml:kml/kml:Document/kml:ExtendedData/"
            + "dwd:ProductDefinition/dwd:IssueTime",
            namespaces=tree.nsmap,
        )[0].text,
        "%Y-%m-%dT%H:%M:%S.%fZ",
    )
    if (datetime.now() - issue_time).total_seconds() > 25200:
        return None

    # Collect all time steps
    timesteps = []
    for step in tree.xpath(
        "//kml:kml/kml:Document/kml:ExtendedData/dwd:ProductDefinition/"
        + "dwd:ForecastTimeSteps/dwd:TimeStep",
        namespaces=tree.nsmap,
    ):
        timesteps.append(step.text)

    # COLLECT WEATHER DATA
    # Each parameter is processed individually
    data = {
        "time": timesteps,
        Parameter.TEMP: [],
        Parameter.DWPT: [],
        Parameter.PRCP: [],
        Parameter.WDIR: [],
        Parameter.WSPD: [],
        Parameter.WPGT: [],
        Parameter.TSUN: [],
        Parameter.PRES: [],
        Parameter.CLDC: [],
        Parameter.VSBY: [],
        Parameter.COCO: [],
    }
    placemark = tree.xpath(
        "//kml:kml/kml:Document/kml:Placemark", namespaces=tree.nsmap
    )[0]

    # Pressure
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="PPPP"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.PRES].append(
            float(value) / 100
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Air temperature
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="TTT"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.TEMP].append(
            kelvin_to_celsius(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Dew point
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="Td"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.DWPT].append(
            kelvin_to_celsius(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Wind direction
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="DD"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.WDIR].append(
            int(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Wind speed
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="FF"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.WSPD].append(
            ms_to_kmh(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Peak wind gust
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="FX1"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.WPGT].append(
            ms_to_kmh(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Weather condition
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="ww"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.COCO].append(
            get_coco(int(float(value)))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Precipitation
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="RR1c"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.PRCP].append(
            float(value) if value.lstrip("-").replace(".", "", 1).isdigit() else None
        )

    # Sunshine Duration
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="SunD1"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.TSUN].append(
            float(value) / 60
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Cloud Cover
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="N"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.CLDC].append(
            percentage_to_okta(float(value))
            if value.lstrip("-").replace(".", "", 1).isdigit()
            else None
        )

    # Visibility
    for value in (
        re.sub(
            r"/\s+/",
            " ",
            placemark.xpath(
                'kml:ExtendedData/dwd:Forecast[@dwd:elementName="VV"]/dwd:value',
                namespaces=tree.nsmap,
            )[0].text,
        )
        .strip()
        .split()
    ):
        data[Parameter.VSBY].append(
            float(value) if value.lstrip("-").replace(".", "", 1).isdigit() else None
        )

    # Convert data dict to DataFrame
    df = pd.DataFrame.from_dict(data)

    # Convert time strings to datetime
    df["time"] = pd.to_datetime(df["time"])

    # Calculate humidity data
    df[Parameter.RHUM] = df.apply(temp_dwpt_to_rhum, axis=1)

    # Set index
    df = df.set_index(["time"])

    # Round decimals
    df = df.round(1)

    # Remove tz awareness
    df = df.tz_convert(None, level="time")

    return df


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if "mosmix" in req.station.identifiers:
        return get_df(req.station.identifiers["mosmix"])



"""
Meteostat Typing
"""

from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Callable, List, Optional

from meteostat.core.validator import Validator
from meteostat.enumerations import (
    Grade,
    Priority,
    Granularity,
    Parameter,
    Provider,
    Unit,
)


@dataclass
class Station:
    """
    A weather station

    For virtual stations created from Point objects, some fields will be None.
    """

    id: str  # The Meteostat station ID (e.g., "10637" or "$0001" for virtual stations)
    name: Optional[str] = None  # The (usually English) name of the station
    country: Optional[str] = None  # ISO 3166-1 alpha-2 country code
    region: Optional[str] = None  # ISO 3166-2 state or region code
    identifiers: dict[str, str] = field(default_factory=dict)  # Provider identifiers
    latitude: Optional[float] = None  # The latitude in degrees
    longitude: Optional[float] = None  # The longitude in degrees
    elevation: Optional[int] = None  # The elevation in meters
    timezone: Optional[str] = None  # The IANA timezone name


@dataclass
class License:
    """
    A license
    """

    commercial: bool
    attribution: Optional[str] = None
    name: Optional[str] = None
    url: Optional[str] = None


@dataclass
class ProviderSpec:
    """
    A provider's meta data
    """

    id: Provider | str  # The provider ID
    name: str  # A descriptive provider name
    granularity: Granularity  # The provider's time series granularity
    priority: Priority | int  # The priority of the provider
    grade: Optional[Grade]  # The provider's data quality grade
    license: Optional[License]  # The provider's license
    parameters: List[Parameter]  # List of supported meteorological parameters
    start: date  # The start date of the provider's data
    end: Optional[datetime] = None  # The end date of the provider's data
    countries: Optional[list[str]] = None  # List of supported countries
    module: Optional[str] = None  # Module path to the provider's API


@dataclass
class ParameterSpec:
    """
    A parameter's meta data
    """

    id: Parameter | str  # The parameter ID
    name: str  # A descriptive parameter name
    granularity: Granularity  # The parameter's granularity
    dtype: str  # The parameter's data type
    unit: Optional[Unit] = None  # The parameter's data unit
    validators: List[Validator | Callable] = field(
        default_factory=list
    )  # The parameter's validators


@dataclass
class Request:
    """
    A request to fetch meteorological time series data
    """

    granularity: Granularity  # Query's time series granularity
    providers: List[Provider]  # Providers to query
    parameters: List[Parameter]  # Schema of the query's data
    station: Station | List[Station]  # Station(s) to query
    start: Optional[datetime] = None  # Start date of the query
    end: Optional[datetime] = None  # End date of the query
    timezone: Optional[str] = None  # Time zone of the query's data


@dataclass
class ProviderRequest:
    """
    A query to fetch meteorological data from a provider
    """

    station: Station  # Station to query
    parameters: List[Parameter]  # List of meteorological parameters to query
    start: Optional[datetime]  # Start date of the query
    end: Optional[datetime]  # End date of the query



"""
Meteostat Enumerations
"""

from enum import StrEnum, IntEnum


class Granularity(StrEnum):
    """
    The different levels of time series granularity
    """

    HOURLY = "hourly"
    DAILY = "daily"
    MONTHLY = "monthly"
    NORMALS = "normals"


class Frequency(StrEnum):
    """
    The different levels of time series frequency
    """

    HOURLY = "h"
    DAILY = "D"
    MONTHLY = "MS"


class Parameter(StrEnum):
    """
    The different meteorological parameters supported by Meteostat
    """

    TEMP = "temp"  # Air temperature (aggregation: mean)
    TMIN = "tmin"  # Daily minimum air temperature (aggregation: mean)
    TMAX = "tmax"  # Daily maximum air temperature (aggregation: mean)
    TXMN = "txmn"  # Extreme minimum air temperature (aggregation: min)
    TXMX = "txmx"  # Extreme maximum air temperature (aggregation: max)
    DWPT = "dwpt"  # Dew point (aggregation: mean)
    PRCP = "prcp"  # Precipitation (aggregation: sum)
    PDAY = "pday"  # Days with precipitation equal to or greater than 1 millimeter (aggregation: sum)
    WDIR = "wdir"  # Wind direction at observation time
    WSPD = "wspd"  # Wind speed (aggregation: mean)
    WPGT = "wpgt"  # Peak wind gust (aggregation: max)
    RHUM = "rhum"  # Relative humidity (aggregation: mean)
    PRES = "pres"  # Air pressure at MSL (aggregation: mean)
    SNWD = "snwd"  # Snow depth on ground
    SNOW = "snow"  # Snowfall (aggregation: sum)
    TSUN = "tsun"  # Sunshine duration (aggregation: sum)
    SGHI = "sghi"  # TBD
    SDNI = "sdni"  # TBD
    SDHI = "sdhi"  # TBD
    CLDC = "cldc"  # Cloud cover (aggregation: mean)
    VSBY = "vsby"  # Visibility (aggregation: mean)
    COCO = "coco"  # Weather condition code at time of observation


class Unit(StrEnum):
    """
    Data Units
    """

    CELSIUS = "°C"
    FAHRENHEIT = "°F"
    KELVIN = "K"
    PERCENTAGE = "%"
    HPA = "hPa"
    MILLIMETERS = "mm"
    CENTIMETERS = "cm"
    METERS = "m"
    KMH = "km/h"
    DEGREES = "°"
    MINUTES = "min"
    OKTAS = "okta"


class UnitSystem(StrEnum):
    """
    Unit Systems
    """

    SI = "si"
    METRIC = "metric"
    IMPERIAL = "imperial"


class Provider(StrEnum):
    """
    Providers supported by Meteostat
    """

    ISD_LITE = "isd_lite"
    METAR = "metar"
    GHCND = "ghcnd"
    CLIMAT = "climat"
    DWD_HOURLY = "dwd_hourly"
    DWD_POI = "dwd_poi"
    DWD_MOSMIX = "dwd_mosmix"
    DWD_DAILY = "dwd_daily"
    DWD_MONTHLY = "dwd_monthly"
    ECCC_HOURLY = "eccc_hourly"
    ECCC_DAILY = "eccc_daily"
    ECCC_MONTHLY = "eccc_monthly"
    METNO_FORECAST = "metno_forecast"
    GSA_HOURLY = "gsa_hourly"
    GSA_SYNOP = "gsa_synop"
    GSA_DAILY = "gsa_daily"
    GSA_MONTHLY = "gsa_monthly"

    HOURLY = "hourly"
    DAILY = "daily"
    DAILY_DERIVED = "daily_derived"
    MONTHLY = "monthly"
    MONTHLY_DERIVED = "monthly_derived"


class Priority(IntEnum):
    """
    Provider priorities
    """

    HIGHEST = 25
    HIGH = 20
    MEDIUM = 15
    LOW = 10
    LOWEST = 5
    NONE = 0


class Grade(IntEnum):
    """
    Provider quality grades
    """

    RECORD = 4
    OBSERVATION = 3
    ANALYSIS = 2
    FORECAST = 1


class TTL(IntEnum):
    """
    Cache TTLs
    """

    HOUR = 60 * 60
    DAY = 60 * 60 * 24
    WEEK = 60 * 60 * 24 * 7
    MONTH = 60 * 60 * 24 * 30



"""
Data parsers

The code is licensed under the MIT license.
"""

import calendar
from typing import List
import datetime

import pandas as pd
import pytz

from meteostat.api.stations import stations as stations_service
from meteostat.api.point import Point
from meteostat.typing import Station


def parse_station(
    station: (str | Station | Point | List[str | Station | Point] | pd.DataFrame),
) -> Station | List[Station]:
    """
    Parse one or multiple station(s) or geo point(s)

    Point objects are converted to virtual stations with IDs like $0001, $0002, etc.
    based on their position in the input list.

    Returns
    -------
    Station | List[Station]
        - Returns a single Station object for single-station input (str, Station, Point)
        - Returns a list of Station objects for multi-station input (list, pd.Index, etc.)
    """
    # Return data if it contains station meta data (single station)
    if isinstance(station, Station):
        return station

    # Handle Point objects (single point)
    if isinstance(station, Point):
        return _point_to_station(station, 1)

    # Handle string (single station ID)
    if isinstance(station, str):
        meta = stations_service.meta(station)
        if meta is None:
            raise ValueError(f'Weather station with ID "{station}" could not be found')
        return meta

    # Convert station identifier(s) to list (multi-station)
    if isinstance(station, pd.DataFrame):
        stations = station.index.tolist()
    else:
        # It's a list
        stations = station

    # Get station meta data
    data = []
    point_counter = 0
    for s in stations:
        # Append data early if it contains station meta data
        if isinstance(s, Station):
            data.append(s)
            continue
        # Handle Point objects
        if isinstance(s, Point):
            point_counter += 1
            data.append(_point_to_station(s, point_counter))
            continue
        # Get station meta data
        meta = stations_service.meta(s)
        # Raise exception if station could not be found
        if meta is None:
            raise ValueError(f'Weather station with ID "{s}" could not be found')
        # Append station meta data
        data.append(meta)

    # Return list of station meta data
    return data


def _point_to_station(point: Point, index: int) -> Station:
    """
    Convert a Point object to a virtual Station object

    Parameters
    ----------
    point : Point
        The Point object to convert
    index : int
        The position in the list of points (1-indexed)

    Returns
    -------
    Station
        A virtual Station object with an ID like $0001
    """
    # Create virtual station ID
    station_id = f"${index:04d}"

    # Create Station object with extracted coordinates
    return Station(
        id=station_id,
        name=f"Location #{index}",
        latitude=point.latitude,
        longitude=point.longitude,
        elevation=point.elevation,
    )


def parse_time(
    value: datetime.date | datetime.datetime | None,
    timezone: str | None = None,
    is_end: bool = False,
) -> datetime.datetime | None:
    """
    Convert a given date/time input to datetime

    To set the time of a date to 23:59:59, pass is_end=True
    """
    if not value:
        return None

    if not isinstance(value, datetime.datetime):
        parsed = datetime.datetime.combine(
            value,
            datetime.datetime.max.time() if is_end else datetime.datetime.min.time(),
        )
    else:
        parsed = value

    if timezone:
        tz = pytz.timezone(timezone)
        # If the datetime is already aware, convert it directly
        # If it's naive, treat it as being in the target timezone
        if parsed.tzinfo is not None:
            parsed = parsed.astimezone(pytz.utc).replace(tzinfo=None)
        else:
            parsed = tz.localize(parsed).astimezone(pytz.utc).replace(tzinfo=None)

    return parsed


def parse_month(
    value: datetime.date | datetime.datetime | None, is_end: bool = False
) -> datetime.date | None:
    """
    Convert a given date/time input to the first or last day of the month respectively
    """
    if not value:
        return None

    last_day = calendar.monthrange(value.year, value.month)[1]

    return datetime.date(value.year, value.month, last_day if is_end else 1)


def parse_year(year: int, is_end: bool = False) -> datetime.date:
    """
    Parse a year into a date, returning either the first or last day of the year
    """
    return datetime.date(year, 12, 31) if is_end else datetime.date(year, 1, 1)



from datetime import datetime
from typing import Optional

import pandas as pd
import pytz

from meteostat.enumerations import TTL, Parameter
from meteostat.core.cache import cache_service
from meteostat.core.network import network_service
from meteostat.providers.eccc.shared import ENDPOINT, get_meta_data
from meteostat.typing import ProviderRequest

BATCH_LIMIT = 9000
PROPERTIES = {
    "UTC_DATE": "time",
    "RELATIVE_HUMIDITY": Parameter.RHUM,
    "WIND_DIRECTION": Parameter.WDIR,
    "WIND_SPEED": Parameter.WSPD,
    "VISIBILITY": Parameter.VSBY,
    "PRECIP_AMOUNT": Parameter.PRCP,
    "TEMP": Parameter.TEMP,
}


@cache_service.cache(TTL.DAY, "pickle")
def get_df(climate_id: str, year: int, tz: str) -> Optional[pd.DataFrame]:
    # Process start & end date
    # ECCC uses the station's local time zone
    from_timezone = pytz.timezone("UTC")
    to_timezone = pytz.timezone(tz)
    start = (
        from_timezone.localize(datetime(year, 1, 1, 0, 0, 0))
        .astimezone(to_timezone)
        .strftime("%Y-%m-%dT%H:%M:%S")
    )
    end = (
        from_timezone.localize(datetime(year, 12, 31, 23, 59, 59))
        .astimezone(to_timezone)
        .strftime("%Y-%m-%dT%H:%M:%S")
    )

    response = network_service.get(
        f"{ENDPOINT}/collections/climate-hourly/items",
        params={
            "CLIMATE_IDENTIFIER": climate_id,
            "datetime": f"{start}/{end}",
            "f": "json",
            "properties": ",".join(PROPERTIES.keys()),
            "limit": BATCH_LIMIT,
        },
    )

    data = response.json()

    # Extract features from the response
    features = map(
        lambda feature: feature["properties"] if "properties" in feature else {},
        data.get("features", []),
    )

    # Create a DataFrame from the extracted features
    df = pd.DataFrame(features)
    df = df.rename(columns=PROPERTIES)

    # Handle time column & set index
    df["time"] = pd.to_datetime(df["time"])
    df = df.set_index(["time"])

    # Convert data units
    df[Parameter.WDIR] = df[Parameter.WDIR] * 10  # Wind direction is provided 10's deg
    df[Parameter.VSBY] = (
        df[Parameter.VSBY] * 1000
    )  # Visibility is provided in kilometres

    return df


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if (
        "national" not in req.station.identifiers
        or req.start is None
        or req.end is None
    ):
        return None

    meta_data = get_meta_data(req.station.identifiers["national"])
    climate_id = meta_data.get("CLIMATE_IDENTIFIER")
    archive_first = meta_data.get("HLY_FIRST_DATE")
    archive_last = meta_data.get("HLY_LAST_DATE")
    timezone = meta_data.get("TIMEZONE")

    if not (climate_id and archive_first and archive_last and timezone):
        return None

    archive_start = datetime.strptime(archive_first, "%Y-%m-%d %H:%M:%S")
    archive_end = datetime.strptime(archive_last, "%Y-%m-%d %H:%M:%S")

    years = range(
        max(req.start.year, archive_start.year),
        min(req.end.year, archive_end.year) + 1,
    )
    data = [get_df(climate_id, year, timezone) for year in years]

    return pd.concat(data) if len(data) and not all(d is None for d in data) else None



from typing import Optional

import numpy as np
import pandas as pd

from meteostat.enumerations import Parameter, Provider
from meteostat.api.hourly import hourly
from meteostat.typing import ProviderRequest
from meteostat.utils.data import aggregate_sources, reshape_by_source


def daily_mean(group: pd.Series):
    """
    Calculate the daily mean from a series of hourly data
    """
    if group.isna().sum() > 3:
        return np.nan
    return group.interpolate(axis=0).mean()


def daily_min(group: pd.Series):
    """
    Calculate the daily minimum from a series of hourly data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.min()


def daily_max(group: pd.Series):
    """
    Calculate the daily maximum from a series of hourly data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.max()


def daily_sum(group: pd.Series):
    """
    Calculate the daily sum from a series of hourly data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.sum()


# Available parameters with source column and aggregation method
PARAMETER_AGGS = {
    Parameter.TEMP: (Parameter.TEMP, daily_mean),
    Parameter.TMIN: (Parameter.TEMP, daily_min),
    Parameter.TMAX: (Parameter.TEMP, daily_max),
    Parameter.RHUM: (Parameter.RHUM, daily_mean),
    Parameter.DWPT: (Parameter.DWPT, daily_mean),
    Parameter.PRCP: (Parameter.PRCP, daily_sum),
    Parameter.SNWD: (Parameter.SNWD, daily_max),
    Parameter.WSPD: (Parameter.WSPD, daily_mean),
    Parameter.WPGT: (Parameter.WPGT, daily_max),
    Parameter.PRES: (Parameter.PRES, daily_mean),
    Parameter.TSUN: (Parameter.TSUN, daily_sum),
    Parameter.CLDC: (Parameter.CLDC, daily_mean),
}


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """
    Fetch hourly weather data from Meteostat's central data
    repository and aggregate to daily granularity
    """
    # Get all source columns
    source_cols = list(dict.fromkeys([PARAMETER_AGGS[p][0] for p in req.parameters]))

    # Get hourly DataFrame
    ts_hourly = hourly(
        req.station.id,
        req.start,
        req.end,
        parameters=source_cols,
        providers=[Provider.HOURLY],
        timezone=req.station.timezone,
    )

    df_hourly = ts_hourly.fetch(fill=True, sources=True)

    # If no hourly data is available, exit
    if df_hourly is None:
        return None

    # Create daily aggregations
    df = pd.DataFrame()
    for parameter in req.parameters:
        [hourly_param_name, agg_func] = PARAMETER_AGGS[parameter]
        df[parameter] = (
            df_hourly[hourly_param_name]
            .groupby(pd.Grouper(level="time", freq="1D"))
            .agg(agg_func)
        )
        df[f"{parameter}_source"] = (
            df_hourly[f"{hourly_param_name}_source"]
            .groupby(pd.Grouper(level="time", freq="1D"))
            .agg(aggregate_sources)
        )

    # Adjust DataFrame and add index
    df = df.round(1)
    dt_index = pd.DatetimeIndex(df.index)
    df.index = dt_index.tz_localize(None).normalize()
    df.index.name = "time"

    # Return reshaped DataFrame
    return reshape_by_source(df)



"""
DataFrame mutations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from collections import Counter
from datetime import datetime
from itertools import chain
from typing import List

import numpy as np
import pandas as pd

from meteostat.core.providers import provider_service
from meteostat.enumerations import Frequency
from meteostat.typing import Station


def stations_to_df(stations: List[Station]) -> pd.DataFrame:
    """
    Convert list of stations to DataFrame
    """
    assert len(stations) > 0 and all(
        isinstance(station, Station) for station in stations
    )

    return pd.DataFrame.from_records(
        [
            {
                "id": station.id,
                "name": station.name,
                "country": station.country,
                "latitude": station.latitude,
                "longitude": station.longitude,
                "elevation": station.elevation,
                "timezone": station.timezone,
            }
            for station in stations
        ],
        index="id",
    )


def squash_df(df: pd.DataFrame, sources=False) -> pd.DataFrame:
    """
    Squash a DataFrame based on the source priority
    """
    # Capture the columns
    columns = df.columns

    # Add source priority column
    df["source_prio"] = df.index.get_level_values("source").map(
        provider_service.get_source_priority
    )

    # Shift source information to columns
    if sources:
        df = df.reset_index(level="source")
        for column in columns:
            df[f"{column}_source"] = np.where(df[column].notna(), df["source"], np.nan)
        df = df.set_index("source", append=True)

    # Get highest priority value/source for each station and time
    df = (
        df.groupby(level=["station", "time", "source"])
        .last()  # In case of duplicate index, the last row will be prefered
        .sort_values(by="source_prio", ascending=False)
        .groupby(["station", "time"])
        .first()  # Prefer value with highest priority
        .drop("source_prio", axis=1)
    )

    # Order columns and return squashed DataFrame
    return df[order_source_columns(df.columns)] if sources else df


def fill_df(
    df: pd.DataFrame, start: datetime, end: datetime, freq: str
) -> pd.DataFrame:
    """
    Fill a DataFrame with a complete date range for each station
    """
    try:
        iterables = [
            df.index.get_level_values("station").unique(),
            pd.date_range(start=start, end=end, freq=freq),
        ]
        names = ["station", "time"]
        if "source" in df.index.names:
            iterables.append(df.index.get_level_values("source").unique())
            names.append("source")
        # Create a new MultiIndex with every hour for each station
        new_index = pd.MultiIndex.from_product(iterables, names=names)

        # Reindex the DataFrame to add missing rows
        df = df.reindex(index=new_index)
        return df
    except KeyError:
        return df


def localize(df: pd.DataFrame, timezone: str) -> pd.DataFrame:
    """
    Convert time data to any time zone
    """
    return df.tz_localize("UTC", level="time").tz_convert(timezone, level="time")


def reshape_by_source(df: pd.DataFrame) -> pd.DataFrame:
    """
    Reshape a DataFrame so that the source columns are pivoted into
    their own level of the index
    """
    # Extract value rows and source rows
    value_rows = df.loc[:, ~df.columns.str.endswith("_source")]
    source_rows = df.loc[:, df.columns.str.endswith("_source")]

    # Melt both value_rows and source_rows
    value_melted = value_rows.reset_index().melt(id_vars="time")
    source_melted = source_rows.reset_index().melt(id_vars="time")

    # Remove '_source' from the variable names in source_melted
    source_melted["variable"] = source_melted["variable"].str.replace("_source", "")

    # Merge the melted DataFrames
    merged_df = pd.merge(
        value_melted, source_melted, on=["time", "variable"], suffixes=("", "_source")
    )

    # Drop rows with missing values
    merged_df = merged_df.dropna(subset=["value"])

    # Pivot the DataFrame
    df_pivoted = merged_df.pivot(
        index=["time", "value_source"], columns="variable", values="value"
    )

    # Flatten the MultiIndex
    df_pivoted.index = df_pivoted.index.rename(["time", "source"])
    df_pivoted.columns.name = None

    # Preserve the original parameter column order from value_rows
    # Pivot often produces alphabetical column order; reindex to match input
    ordered_columns = [col for col in value_rows.columns if col in df_pivoted.columns]
    if ordered_columns:
        df_pivoted = df_pivoted.reindex(columns=ordered_columns)

    return df_pivoted


def aggregate_sources(series: pd.Series) -> str:
    """
    Concatenate multiple data sources into a unique source string,
    ordered by the number of occurrences.
    """
    # Extract sources and flatten them
    sources = [str(item) for item in series if pd.notna(item)]
    flat_sources = list(chain(*[source.split() for source in sources]))

    # Count occurrences of each source
    source_counts = Counter(flat_sources)

    # Sort sources by count in descending order
    sorted_sources = sorted(source_counts, key=lambda s: source_counts[s], reverse=True)

    # Concatenate sorted sources into a unique source string
    return " ".join(sorted_sources)


def enforce_freq(df: pd.DataFrame, freq: Frequency) -> pd.DataFrame:
    """
    Enforce a specific frequency on a DataFrame by resampling
    """
    df.index = pd.to_datetime(df.index.get_level_values("time"))
    return df.resample(freq).first()


def order_source_columns(columns: pd.Index) -> List[str]:
    """
    Order source columns
    """
    ordered_columns = []

    for col in columns:
        if not col.endswith("_source"):
            ordered_columns.append(col)
            ordered_columns.append(f"{col}_source")

    return ordered_columns



"""
Schema Column Validator

This class is used to validate DataFrame columns based
on a parameter specification.
"""

from inspect import signature
from typing import Callable

from pandas import DataFrame, Series


class Validator:
    """
    Schema Column Validator
    """

    func: Callable
    ignore_na = False
    is_relational = False

    def __init__(self, func: Callable, ignore_na=False, is_relational=False):
        self.func = func
        self.ignore_na = ignore_na
        self.is_relational = is_relational

    def test(self, series: Series, df: DataFrame, column: str) -> bool | Series:
        """
        Run validator

        Returns a bool series:
        True -> Check passed
        False -> Check failed
        """
        arg_count = len((signature(self.func)).parameters)
        args = [series, df, column]
        return self.func(*args[0:arg_count])



"""
Utilities for geographical calculations.

The code is licensed under the MIT license.
"""

import numpy as np


def get_distance(lat1, lon1, lat2, lon2) -> int:
    """
    Calculate distance between two geographical points using the Haversine formula
    """
    # Earth radius in meters
    radius = 6371000

    # Degress to radian
    lat1, lon1, lat2, lon2 = map(np.deg2rad, [lat1, lon1, lat2, lon2])

    # Deltas
    dlat = lat2 - lat1
    dlon = lon2 - lon1

    # Calculate distance
    arch = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    arch_sin = 2 * np.arcsin(np.sqrt(arch))

    return round(radius * arch_sin)



"""
DWD national hourly data import routine

Get hourly data for weather stations in Germany.

The code is licensed under the MIT license.
"""

from datetime import datetime
from ftplib import FTP
from io import BytesIO
from typing import Callable, Dict, List, NotRequired, Optional, TypedDict
from zipfile import ZipFile

import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.core.logger import logger
from meteostat.typing import ProviderRequest, Station
from meteostat.core.cache import cache_service
from meteostat.api.config import config
from meteostat.utils.conversions import ms_to_kmh
from meteostat.providers.dwd.shared import get_condicode
from meteostat.providers.dwd.shared import get_ftp_connection


class ParameterDefinition(TypedDict):
    dir: str
    usecols: List[int]
    names: Dict[str, str]
    convert: NotRequired[Dict[str, Callable]]
    encoding: NotRequired[str]
    historical_only: NotRequired[bool]


BASE_DIR = "/climate_environment/CDC/observations_germany/climate/hourly/"
PARAMETERS: List[ParameterDefinition] = [
    {
        "dir": "precipitation",
        "usecols": [1, 3],
        "names": {"R1": Parameter.PRCP},
    },
    {
        "dir": "air_temperature",
        "usecols": [1, 3, 4],
        "names": {"TT_TU": Parameter.TEMP, "RF_TU": Parameter.RHUM},
    },
    {
        "dir": "wind",
        "usecols": [1, 3, 4],
        "names": {"F": Parameter.WSPD, "D": Parameter.WDIR},
        "convert": {"wspd": ms_to_kmh},
    },
    {
        "dir": "pressure",
        "usecols": [1, 3],
        "names": {"P": Parameter.PRES},
    },
    {
        "dir": "sun",
        "usecols": [1, 3],
        "names": {"SD_SO": Parameter.TSUN},
    },
    {
        "dir": "cloudiness",
        "usecols": [1, 4],
        "names": {"V_N": Parameter.CLDC},
    },
    {
        "dir": "visibility",
        "usecols": [1, 4],
        "names": {"V_VV": Parameter.VSBY},
    },
    {
        "dir": "weather_phenomena",
        "usecols": [1, 3],
        "names": {"WW": Parameter.COCO},
        "convert": {"coco": get_condicode},
        "encoding": "latin-1",
    },
    # TODO: Implement solar radiation
    # {
    #     "dir": "solar",
    #     "usecols": [1, 5],
    #     "names": {"FG_LBERG": "srad"},
    #     "convert": {"srad": jcm2_to_wm2},
    #     "historical_only": True,
    # },
]


def find_file(ftp: FTP, path: str, needle: str):
    """
    Find file in directory
    """
    match = None

    try:
        ftp.cwd(BASE_DIR + path)
        files = ftp.nlst()
        matching = [f for f in files if needle in f]
        match = matching[0]
        logger.debug(f"Found file '{match}' in '{path}' directory")
    except IndexError:
        logger.info(f"File '{needle}' not found in '{path}' directory")

    return match


@cache_service.cache(TTL.DAY, "pickle")
def get_df(parameter_dir: str, mode: str, station_id: str) -> Optional[pd.DataFrame]:
    """
    Get a file from DWD FTP server and convert to Polars DataFrame
    """
    logger.debug(
        f"Fetching {parameter_dir} data ({mode}) for DWD station '{station_id}'"
    )

    parameter = next(param for param in PARAMETERS if param["dir"] == parameter_dir)

    ftp = get_ftp_connection()
    remote_file = find_file(ftp, f"{parameter['dir']}/{mode}", station_id)

    if remote_file is None:
        return None

    buffer = BytesIO()
    ftp.retrbinary("RETR " + remote_file, buffer.write)

    # Unzip file
    with ZipFile(buffer, "r") as zipped:
        filelist = zipped.namelist()
        raw = None
        for file in filelist:
            if file[:7] == "produkt":
                with zipped.open(file, "r") as reader:
                    raw = BytesIO(reader.read())

    # Convert raw data to DataFrame
    df: pd.DataFrame = pd.read_csv(  # type: ignore
        raw,
        sep=";",
        skipinitialspace=True,
        date_format="%Y%m%d%H",
        na_values=[-999, "-999"],
        usecols=parameter["usecols"],
        encoding=parameter["encoding"] if "encoding" in parameter else None,
    )

    df["time"] = pd.to_datetime(df.pop("MESS_DATUM"), format="%Y%m%d%H")

    logger.debug(f"Found {len(df)} rows in {remote_file}")

    # Rename columns
    df = df.rename(columns=lambda x: x.strip())
    df = df.rename(columns=parameter["names"])

    # Convert column data
    if "convert" in parameter:
        for col, func in parameter["convert"].items():
            df[col] = df[col].apply(func)

    # Set index
    df = df.set_index("time")

    # Round decimals
    df = df.round(1)

    return df


def get_parameter(
    parameter_dir: str, modes: list[str], station: Station
) -> Optional[pd.DataFrame]:
    logger.debug(f"Fetching {parameter_dir} data ({modes}) for station '{station.id}'")
    try:
        data = [
            get_df(parameter_dir, mode, station.identifiers["national"])
            for mode in modes
        ]
        if all(d is None for d in data):
            return None
        df = pd.concat(data)
        return df.loc[~df.index.duplicated(keep="first")]
    except Exception as error:
        logger.warning(error, exc_info=True)
        return None


def fetch(req: ProviderRequest):
    if "national" not in req.station.identifiers:
        return None

    # Check which modes to consider for data fetching
    #
    # The dataset is divided into a versioned part with completed quality check ("historical"),
    # and a part for which the quality check has not yet been completed ("recent").
    #
    # There is no definite answer as to when the quality check is completed. We're assuming a
    # period of 3 years here. If the end date of the query is within this period, we will also
    # consider the "recent" mode.
    modes = ["historical"]
    if abs((req.end - datetime.now()).days) < 3 * 365:
        modes.append("recent")

    columns = map(
        lambda args: get_parameter(*args),
        (
            (parameter["dir"], config.dwd_hourly_modes or modes, req.station)
            for parameter in [
                param
                for param in PARAMETERS
                if not set(req.parameters).isdisjoint(param["names"].values())
            ]
        ),
    )

    return pd.concat(columns, axis=1)



"""
Concatenation Module

Provides functions to concatenate multiple time series objects into one.
"""

from copy import copy
from datetime import datetime
from typing import List, Optional

import pandas as pd

from meteostat.core.data import data_service
from meteostat.core.schema import schema_service
from meteostat.api.timeseries import TimeSeries


def _get_dt(
    dt_a: Optional[datetime], dt_b: Optional[datetime], start=True
) -> Optional[datetime]:
    """
    Return the earlier or later (depending on "start" argument) of two datetimes,
    considering None as 'no value'.

    If both are None, return None.
    """
    if dt_a is None:
        return dt_b
    if dt_b is None:
        return dt_a
    return min(dt_a, dt_b) if start else max(dt_a, dt_b)


def merge(objs: List[TimeSeries]) -> TimeSeries:
    """
    Merge one or multiple Meteostat time series into a common one

    In case of duplicate index, the last row will be prefered.
    Hence, please pass newest data last.

    Parameters
    ----------
    objs : List[TimeSeries]
        List of time series objects to concatenate

    Returns
    -------
    TimeSeries
        Concatenated time series object

    Raises
    ------
    ValueError
        If the time series objects have divergent granularity or time zone
    """
    ts = objs[0]

    if not all(
        obj.granularity == ts.granularity and obj.timezone == ts.timezone
        for obj in objs[1:]
    ):
        raise ValueError(
            "Can't concatenate time series objects with divergent granularity or time zone"
        )

    stations = copy(ts.stations)
    start = copy(ts.start)
    end = copy(ts.end)
    parameters = ts.parameters
    multi_station = ts._multi_station

    for obj in objs[1:]:
        stations = (
            pd.concat([stations, obj.stations])
            .reset_index()
            .drop_duplicates(subset=["id"])
            .set_index("id")
        )
        start = _get_dt(start, obj.start)
        end = _get_dt(end, obj.end, False)
        parameters.extend(obj.parameters)
        if (
            obj._multi_station
            or stations.index.get_level_values("id")[0]
            != obj.stations.index.get_level_values("id")[0]
        ):
            multi_station = True

    df = data_service.concat_fragments(
        [obj._df for obj in objs if obj._df is not None],
        list(dict.fromkeys(parameters)),
    )
    df = schema_service.format(df, ts.granularity)

    return TimeSeries(
        ts.granularity,
        stations,
        df,
        start,
        end,
        ts.timezone,
        multi_station=multi_station,
    )



"""
Provider Service

The Provider Service provides methods to interact with data providers.
"""

from datetime import datetime
from importlib import import_module
from statistics import fmean
from typing import List, Optional, TypeGuard, cast

import pandas as pd

from meteostat.core.logger import logger
from meteostat.enumerations import Granularity, Priority, Provider
from meteostat.providers.index import DEFAULT_PROVIDERS
from meteostat.typing import (
    ProviderRequest,
    ProviderSpec,
    Station,
    Request,
)


class ProviderService:
    """
    Provider Service
    """

    _providers: List[ProviderSpec]

    def __init__(self, providers: List[ProviderSpec]) -> None:
        self._providers = providers

    @property
    def providers(self) -> List[ProviderSpec]:
        """
        Get supported providers
        """
        return self._providers

    def register(self, provider: ProviderSpec) -> None:
        """
        Register a provider
        """
        self._providers.append(provider)

    def get_provider(self, provider_id: Provider | str) -> Optional[ProviderSpec]:
        """
        Get provider by ID
        """
        return next(
            (provider for provider in self._providers if provider.id == provider_id),
            None,
        )

    def _get_provider_priority(self, provider_id: Provider | str) -> int:
        """
        Get priority of a provider by its ID
        """
        baselines = {
            Granularity.HOURLY: 0,
            Granularity.DAILY: 100,
            Granularity.MONTHLY: 200,
            Granularity.NORMALS: 300,
        }

        provider = self.get_provider(provider_id)

        if not provider:
            return Priority.NONE

        baseline = baselines[provider.granularity]

        return int(provider.priority + baseline)

    def get_source_priority(self, source: str) -> float:
        """
        Get priority of a source string
        """
        provider_ids = source.split(" ")

        if len(provider_ids) == 1:
            return self._get_provider_priority(provider_ids[0])

        priorities = [
            self._get_provider_priority(provider) for provider in provider_ids
        ]

        return fmean(priorities)

    def filter_providers(self, query: Request, station: Station) -> List[Provider]:
        """
        Get a filtered list of providers
        """

        def _filter(provider_id: Provider) -> TypeGuard[Provider]:
            provider = self.get_provider(provider_id)

            if provider is None:
                return False

            # Filter out providers with diverging granularities
            if provider.granularity is not query.granularity:
                logger.error(
                    "Provider '%s' does not support granularity '%s'",
                    provider_id,
                    query.granularity,
                )
                return False

            # Filter out providers with no overlap in parameters
            if set(provider.parameters).isdisjoint(query.parameters):
                logger.info(
                    "Provider '%s' does not support any requested parameter",
                    provider_id,
                )
                return False

            # Filter out providers which do not serve the station's country
            if provider.countries and station.country not in provider.countries:
                logger.info(
                    "Skipping provider '%s' as it does not serve the station's country ('%s')",
                    provider_id,
                    station.country,
                )
                return False

            # Filter out providers which stopped providing data before the request's start date
            if query.end and query.end < datetime.combine(
                provider.start, datetime.min.time()
            ):
                logger.info(
                    "Skipping provider '%s' as it stopped providing data before request start",
                    provider_id,
                )
                return False

            # Filter out providers which only started providing data after the request's end date
            if (
                provider.end is not None
                and query.start is not None
                and query.start > datetime.combine(provider.end, datetime.max.time())
            ):
                logger.info(
                    "Skipping provider '%s' as it only started providing data after request end",
                    provider_id,
                )
                return False

            return True

        return list(filter(_filter, query.providers))

    def fetch_data(
        self, provider_id: Provider, req: Request, station: Station
    ) -> Optional[pd.DataFrame]:
        """
        Fetch data from a given provider
        """
        provider = self.get_provider(provider_id)

        if not provider:
            return None

        query = ProviderRequest(
            station=station,
            start=req.start
            or (
                datetime.combine(provider.start, datetime.min.time())
                if provider.start
                else None
            ),
            end=req.end or (provider.end or datetime.now()),
            parameters=req.parameters,
        )

        module = import_module(cast(str, provider.module))
        df = module.fetch(query)

        return df


provider_service = ProviderService(providers=DEFAULT_PROVIDERS)



"""
The code is licensed under the MIT license.
"""

from typing import Optional

import pandas as pd

from meteostat.api.config import config
from meteostat.enumerations import TTL
from meteostat.providers.meteostat.shared import filter_model_data, handle_exceptions
from meteostat.typing import ProviderRequest
from meteostat.core.cache import cache_service
from meteostat.utils.data import reshape_by_source

ENDPOINT = config.monthly_endpoint


@cache_service.cache(TTL.MONTH, "pickle")
@handle_exceptions
def get_df(station: str) -> Optional[pd.DataFrame]:
    """
    Get CSV file from Meteostat and convert to DataFrame
    """
    file_url = ENDPOINT.format(station=station)

    df = pd.read_csv(file_url, sep=",", compression="gzip")

    time_cols = df.columns[0:2]
    df["date"] = (
        df["year"].astype(str) + "-" + df["month"].astype(str).str.zfill(2) + "-01"
    )
    df["time"] = pd.to_datetime(df["date"])
    df = df.drop(time_cols, axis=1).drop("date", axis=1).set_index("time")

    return reshape_by_source(df)


@filter_model_data
def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """
    Fetch monthly weather data from Meteostat's central data repository
    """
    df = get_df(req.station.id)
    return df



"""
TimeSeries Class

A class to handle meteorological time series data.
"""

from copy import copy
from datetime import datetime
from itertools import chain
from math import floor
from statistics import mean
from typing import List, Optional

import pandas as pd

from meteostat.core.parameters import parameter_service
from meteostat.core.validator import Validator
from meteostat.core.providers import provider_service
from meteostat.core.schema import schema_service
from meteostat.enumerations import Parameter, Granularity, Provider, UnitSystem
from meteostat.typing import License
from meteostat.utils.data import fill_df, localize, squash_df


class TimeSeries:
    """
    TimeSeries class which provides features which are
    used across all granularities
    """

    granularity: Granularity
    stations: pd.DataFrame
    start: Optional[datetime] = None
    end: Optional[datetime] = None
    timezone: Optional[str] = None

    _df: Optional[pd.DataFrame] = None
    _multi_station: bool = False

    def __init__(
        self,
        granularity: Granularity,
        stations: pd.DataFrame,
        df: Optional[pd.DataFrame],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        timezone: Optional[str] = None,
        multi_station: bool = False,
    ) -> None:
        self.granularity = granularity
        self.stations = stations
        self.timezone = timezone
        self._multi_station = multi_station
        if df is not None and not df.empty:
            self._df = df
            self.start = start if start else df.index.get_level_values("time").min()
            self.end = end if end else df.index.get_level_values("time").max()

    def __len__(self) -> int:
        """
        Return number of rows in DataFrame
        """
        return len(self._df) if self._df is not None else 0

    def __str__(self) -> str:
        """
        Return a stringified version of the DataFrame
        """
        return self._df.__str__() if self._df is not None else "Empty time series"

    @property
    def _target_length(self) -> int:
        """
        Expected number of non-NaN values
        """
        if not self.start or not self.end:
            return 0

        diff = self.end - self.start

        return (
            diff.days + 1
            if self.granularity is Granularity.DAILY
            else floor(diff.total_seconds() / 3600) + 1
        ) * len(self.stations)

    @property
    def parameters(self) -> List[Parameter]:
        """
        Get parameters
        """
        return self._df.columns.to_list() if self._df is not None else []

    @property
    def freq(self) -> Optional[str]:
        """
        The time series frequency.

        Returns a Pandas offset alias string (e.g. "1h", "1D", "1M") or
        `None` for granularities where a regular frequency does not apply
        (e.g. normals).
        """
        if self.granularity is Granularity.HOURLY:
            return "1h"

        if self.granularity is Granularity.DAILY:
            return "1D"

        if self.granularity is Granularity.MONTHLY:
            return "1MS"

        # Normals (climatological normals) do not have a regular time
        # frequency in the same sense as timeseries data.
        if self.granularity is Granularity.NORMALS:
            return None

        return None

    @property
    def empty(self) -> bool:
        """
        Is the time series empty?
        """
        return True if self._df is None else self._df.empty

    @property
    def providers(self) -> List[Provider]:
        """
        Get included providers
        """
        if self._df is None:
            return []
        providers: List[str] = (
            self._df.index.get_level_values("source").unique().to_list()
        )
        return list(
            set(chain.from_iterable([provider.split(" ") for provider in providers]))
        )

    @property
    def licenses(self) -> List[License]:
        """
        Get licenses
        """
        providers = [
            provider
            for provider_id in self.providers
            if (provider := provider_service.get_provider(provider_id)) is not None
        ]

        return [
            provider.license for provider in providers if provider.license is not None
        ]

    @property
    def attribution(self) -> str:
        """
        Attribution string
        """
        attributions = [
            "Meteostat",
            *set(
                [
                    license.attribution
                    for license in self.licenses
                    if license.attribution
                ]
            ),
        ]

        return ", ".join(attributions)

    @property
    def commercial(self) -> bool:
        """
        Is commercial use allowed?
        """
        return all(license.commercial for license in self.licenses)

    def fetch(
        self,
        squash=True,
        fill=False,
        sources=False,
        location=False,
        clean=True,
        humanize=False,
        units: UnitSystem = UnitSystem.METRIC,
    ) -> Optional[pd.DataFrame]:
        """
        Fetch the time series data as a DataFrame.

        Parameters
        ----------
        squash : bool, optional
            Whether to squash the DataFrame by source. Defaults to True.
        fill : bool, optional
            Whether to fill missing rows in the DataFrame. Defaults to False.
        sources : bool, optional
            Whether to include source information in the DataFrame. Defaults to False.
        location : bool, optional
            Whether to include location information (latitude, longitude, elevation)
            in the DataFrame. Defaults to False.
        clean : bool, optional
            Whether to clean the DataFrame according to the schema. Defaults to True.
        humanize : bool, optional
            Whether to convert wind direction and condition codes to human-readable values. Defaults to False.
        units : UnitSystem, optional
            The unit system to use for the DataFrame. Defaults to metric units.

        Returns
        -------
        pd.DataFrame or None
            The time series data as a DataFrame, or None if no data is available.
        """
        df = copy(self._df)

        if df is None:
            return None

        if squash:
            df = squash_df(df, sources=sources)

        if clean:
            df = schema_service.clean(df, self.granularity)

        if (
            fill
            and self.start is not None
            and self.end is not None
            and self.freq is not None
        ):
            df = fill_df(df, self.start, self.end, self.freq)

        if self.timezone:
            df = localize(df, self.timezone)

        if location:
            df = df.join(
                self.stations[["latitude", "longitude", "elevation"]], on="station"
            )

        if humanize:
            df = schema_service.humanize(df)

        if units != UnitSystem.METRIC:
            df = schema_service.convert(df, self.granularity, units)

        # Remove station index level if not a multi-station query
        if not self._multi_station and "station" in df.index.names:
            df = df.droplevel("station")

        return df.sort_index()

    def count(self, parameter: Optional[Parameter | str] = None) -> int:
        """
        Get number of non-NaN values for a specific parameter.
        If no parameter is specified, it returns the count for the entire DataFrame.

        Parameters
        ----------
        parameter : Parameter or str
            The parameter to count non-NaN values for. If None, counts for the entire DataFrame.

        Returns
        -------
        int
            The count of non-NaN values for the specified parameter or the entire DataFrame.
        """
        if self._df is None:
            return 0

        if parameter is None:
            return self._df.count().max()

        return self._df[
            parameter if isinstance(parameter, Parameter) else parameter
        ].count()

    def completeness(self, parameter: Optional[Parameter | str] = None) -> float:
        """
        Get completeness for a specific parameter or the entire DataFrame.

        Parameters
        ----------
        parameter : Parameter or str, optional
            The parameter to calculate completeness for.
            If None, calculates for the entire DataFrame.

        Returns
        -------
        float
            The completeness ratio for the specified parameter or the entire DataFrame.
            Returns 0 if no data is available, 1 if complete, or a value between 0 and 1 otherwise.
        """
        df = self.fetch()

        if df is None:
            return 0

        if parameter:
            return round(
                self.count(parameter) / self._target_length,
                2,
            )

        return round(mean([self.completeness(p) for p in df.columns]), 2)

    def validate(self) -> bool:
        """
        Does the time series pass all validations?
        """
        df = self.fetch(fill=True, clean=False)

        if df is None:
            return True

        for col in df:
            parameter = parameter_service.get_parameter(col, self.granularity)
            if parameter is None or not hasattr(parameter, "validators"):
                continue
            for validator in parameter.validators:
                if isinstance(validator, Validator):
                    test = validator.test(df[col], df, col)
                else:
                    test = validator(df[col], df, col)

                if isinstance(test, bool):
                    if not test:
                        return False
                elif not test.all():
                    return False

        return True



"""
DWD national daily data import routine

Get daily data for weather stations in Germany.

The code is licensed under the MIT license.
"""

from datetime import datetime
from ftplib import FTP
from io import BytesIO
from typing import Optional
from zipfile import ZipFile

import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.typing import ProviderRequest
from meteostat.core.cache import cache_service
from meteostat.utils.conversions import ms_to_kmh
from meteostat.providers.dwd.shared import get_ftp_connection


BASE_DIR = "/climate_environment/CDC/observations_germany/climate/monthly/kl/"
USECOLS = [1, 4, 5, 6, 7, 9, 10, 11, 12, 14]  # CSV cols which should be read
NAMES = {
    "MO_N": Parameter.CLDC,
    "MO_TT": Parameter.TEMP,
    "MO_TX": Parameter.TMAX,
    "MO_TN": Parameter.TMIN,
    "MX_TX": Parameter.TXMX,
    "MX_TN": Parameter.TXMN,
    "MX_FX": Parameter.WPGT,
    "MO_SD_S": Parameter.TSUN,
    "MO_RR": Parameter.PRCP,
}


def find_file(ftp: FTP, mode: str, needle: str):
    """
    Find file in directory
    """
    match = None

    try:
        ftp.cwd(BASE_DIR + mode)
        files = ftp.nlst()
        matching = [f for f in files if needle in f]
        match = matching[0]
    except BaseException:
        pass

    return match


@cache_service.cache(TTL.WEEK, "pickle")
def get_df(station: str, mode: str) -> Optional[pd.DataFrame]:
    """
    Get a file from DWD FTP server and convert to Polars DataFrame
    """
    ftp = get_ftp_connection()
    remote_file = find_file(ftp, mode, f"_{station}_")

    if remote_file is None:
        return None

    buffer = BytesIO()
    ftp.retrbinary("RETR " + remote_file, buffer.write)

    ftp.close()

    # Unzip file
    with ZipFile(buffer, "r") as zipped:
        filelist = zipped.namelist()
        raw = None
        for file in filelist:
            if file[:7] == "produkt":
                with zipped.open(file, "r") as reader:
                    raw = BytesIO(reader.read())

    # Convert raw data to DataFrame
    df: pd.DataFrame = pd.read_csv(  # type: ignore
        raw,
        sep=r"\s*;\s*",
        na_values=["-999", -999],
        usecols=USECOLS,
        engine="python",
    )

    # Rename columns
    df = df.rename(columns=lambda x: x.strip())

    # Parse date column
    df["MESS_DATUM_BEGINN"] = pd.to_datetime(
        df["MESS_DATUM_BEGINN"].astype(str), format="%Y%m%d"
    )
    df = df.rename(columns={"MESS_DATUM_BEGINN": "time"})
    df = df.rename(columns=NAMES)

    # Convert data
    df["wpgt"] = df["wpgt"].apply(ms_to_kmh)
    df["tsun"] = df["tsun"] * 60
    df["tsun"] = df["tsun"].round()
    df["cldc"] = df["cldc"].round()

    # Set index
    df = df.set_index("time")

    # Round decimals
    df = df.round(1)

    return df


def fetch(req: ProviderRequest):
    if "national" not in req.station.identifiers:
        return pd.DataFrame()

    # Check which modes to consider for data fetching
    #
    # The dataset is divided into a versioned part with completed quality check ("historical"),
    # and a part for which the quality check has not yet been completed ("recent").
    #
    # There is no definite answer as to when the quality check is completed. We're assuming a
    # period of 3 years here. If the end date of the query is within this period, we will also
    # consider the "recent" mode.
    modes = ["historical"]
    if abs((req.end - datetime.now()).days) < 3 * 365:
        modes.append("recent")

    data = [
        get_df(
            req.station.identifiers["national"],
            mode,
        )
        for mode in modes
    ]

    df = pd.concat(data)

    return df.loc[~df.index.duplicated(keep="first")]



from itertools import combinations
from statistics import mean

import numpy as np
import pandas as pd

from meteostat.api.timeseries import TimeSeries
from meteostat.api.config import config
from meteostat.enumerations import Parameter


def lapse_rate(ts: TimeSeries, parameter: Parameter = Parameter.TEMP) -> float | None:
    """
    Calculate the lapse rate (temperature gradient) in degrees Celsius per kilometer
    based on temperature and elevation data from multiple stations.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing temperature and elevation data for multiple stations.

    Returns
    -------
    float
        Calculated lapse rate in degrees Celsius per kilometer.
    """
    df = ts.fetch(location=True)

    if df is None or "elevation" not in df.columns or parameter not in df.columns:
        return None

    elev_by_station = df["elevation"].groupby(level="station").first()
    temp_by_station = df[parameter].groupby(level="station").mean()

    if len(elev_by_station) < 2 or len(temp_by_station) < 2:
        return None

    lapse_rates = []

    for a, b in combinations(elev_by_station.index, 2):
        if (
            pd.isna(elev_by_station[a])
            or pd.isna(elev_by_station[b])
            or pd.isna(temp_by_station[a])
            or pd.isna(temp_by_station[b])
            or elev_by_station[a] == elev_by_station[b]
        ):
            continue

        temp_diff = temp_by_station[a] - temp_by_station[b]
        elev_diff = elev_by_station[a] - elev_by_station[b]

        # multiply by -1 to get positive lapse rate for decreasing temp
        # with increasing elevation
        lapse_rate = (temp_diff / elev_diff) * 1000 * -1
        lapse_rates.append(lapse_rate)

    if not lapse_rates:
        return None

    return mean(lapse_rates)


def apply_lapse_rate(
    df: pd.DataFrame, elevation: int, lapse_rate: float
) -> pd.DataFrame:
    """
    Calculate approximate temperature at target elevation
    using a given lapse rate.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing the data to be adjusted.
    elevation : int
        Target elevation in meters.
    lapse_rate : float
        Lapse rate (temperature gradient) in degrees Celsius per kilometer.

    Returns
    -------
    pd.DataFrame
        DataFrame with adjusted temperature values.
    """
    for col in config.lapse_rate_parameters:
        if col in df.columns:
            df.loc[df[col] != np.nan, col] = round(
                df[col] + ((lapse_rate / 1000) * (df["elevation"] - elevation)), 1
            )

    return df



"""
Cache Service

The Cache Service provides utilities for caching data on the local file system.
"""

from functools import wraps
from hashlib import md5
import json
import os
from os.path import exists
from time import time
from typing import Any, Callable, Optional

import pandas as pd

from meteostat.api.config import config
from meteostat.core.logger import logger


class CacheService:
    """
    Cache Service
    """

    _purged = False  # Flag to indicate if cache has been purged automatically

    @staticmethod
    def _write_pickle(path: str, df: Optional[pd.DataFrame]) -> None:
        """
        Persist a DataFrame in Pickle format
        """
        if df is None:
            pd.DataFrame().to_pickle(path)
        else:
            df.to_pickle(path)

    @staticmethod
    def _read_pickle(path) -> Optional[pd.DataFrame]:
        """
        Read a pickle file into a DataFrame
        """
        df: pd.DataFrame = pd.read_pickle(path)
        return None if df.empty else df

    @staticmethod
    def _write_json(path: str, data: dict | list) -> None:
        """
        Persist data in JSON format
        """
        with open(path, "w", encoding="utf-8") as file:
            json.dump(data, file)

    @staticmethod
    def _read_json(path) -> dict | list:
        """
        Read JSON data into memory
        """
        with open(path, "r", encoding="utf-8") as file:
            raw = file.read()
        return json.loads(raw)

    @staticmethod
    def _func_to_uid(func, args: tuple, kwargs: dict[str, Any]) -> str:
        """
        Get a unique ID from a function call based on its module, name and arguments
        """
        return md5(
            ";".join(
                (
                    func.__module__,
                    func.__name__,
                    *map(str, args),
                    *[f"{key}:{str(value)}" for key, value in kwargs.items()],
                )
            ).encode("utf-8")
        ).hexdigest()

    @staticmethod
    def create_cache_dir() -> None:
        """
        Create the cache directory if it doesn't exist
        """
        cache_dir = config.cache_directory

        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

    @staticmethod
    def get_cache_path(uid: str, filetype: str):
        """
        Get path of a cached file based on its uid and file type
        """
        return config.cache_directory + os.sep + f"{uid}.{filetype}"

    @staticmethod
    def is_stale(path: str, ttl: int) -> bool:
        """
        Check if a cached file is stale based on its age and TTL
        """
        return time() - os.path.getmtime(path) > max([ttl, config.cache_ttl])

    @staticmethod
    def purge(ttl: Optional[int] = None) -> None:
        """
        Remove stale files from disk cache
        """
        if ttl is None:
            ttl = config.cache_ttl

        logger.debug("Removing cached files older than %s seconds", ttl)

        cache_dir = config.cache_directory

        if os.path.exists(cache_dir):
            # Get current time
            now = time()
            # Go through all files
            for file in os.listdir(cache_dir):
                # Get full path
                path = os.path.join(cache_dir, file)
                # Check if file is older than TTL
                if now - os.path.getmtime(path) > ttl and os.path.isfile(path):
                    # Delete file
                    os.remove(path)

    def persist(
        self, path: str, data: pd.DataFrame | dict | list, data_type: str
    ) -> None:
        """
        Persist any given data under a specific path
        """
        # Create cache directory if it doesn't exist
        self.create_cache_dir()
        # Save data locally
        if data_type == "json" and isinstance(data, (dict, list)):
            self._write_json(path, data)
        elif data_type == "pickle" and (isinstance(data, pd.DataFrame) or data is None):
            self._write_pickle(path, data)

    def fetch(self, path, data_type: str) -> pd.DataFrame | dict | list | None:
        """
        Fetch data from a given path
        """
        if data_type == "json":
            return self._read_json(path)
        return self._read_pickle(path)

    def from_func(
        self, func, args, kwargs, ttl: int, data_format: str
    ) -> pd.DataFrame | dict | list:
        """
        Cache a function's return value
        """
        uid = self._func_to_uid(func, args, kwargs)  # Get UID for function call
        path = self.get_cache_path(uid, data_format)  # Get the local cache path
        result = (
            self.fetch(path, data_format)
            if ttl > 0 and exists(path) and not self.is_stale(path, ttl)
            else False
        )

        cache_status = "is" if isinstance(result, pd.DataFrame) or result else "is not"
        logger.debug(
            "%s from module %s with args=%s and kwargs=%s returns %s and %s served from cache",
            func.__name__,
            func.__module__,
            args,
            kwargs,
            data_format,
            cache_status,
        )

        if isinstance(result, pd.DataFrame) or result:
            return result

        result = func(*args, **kwargs)
        if ttl > 0:
            self.persist(path, result, data_format)

        return result

    def cache(
        self,
        ttl: int | Callable[..., int] = 60 * 60 * 24,
        data_format: str = "json",
    ):
        """
        A simple decorator which caches a function's return value
        based on its payload.

        All data is persisted in either JSON or Pickle format.
        """

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if not config.cache_enable:
                    logger.debug(
                        "Omitting cache for %s from module %s with args=%s and kwargs=%s",
                        func.__name__,
                        func.__module__,
                        args,
                        kwargs,
                    )
                    return func(*args, **kwargs)
                if config.cache_autoclean and not self._purged:
                    self.purge()
                    self._purged = True
                return self.from_func(
                    func,
                    args,
                    kwargs,
                    ttl if isinstance(ttl, int) else ttl(*args, **kwargs),
                    data_format,
                )

            return wrapper

        return decorator


cache_service = CacheService()
purge = cache_service.purge



"""
Point Class

A geographical point used for querying nearby weather stations and spatial interpolation.
"""

from typing import Optional


class Point:
    """
    A geographical point
    """

    latitude: float
    longitude: float
    elevation: Optional[int]

    def __init__(
        self, latitude: float, longitude: float, elevation: Optional[int] = None
    ) -> None:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90")

        if longitude < -180 or longitude > 180:
            raise ValueError("Longitude must be between -180 and 180")

        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation



"""
Meteostat Parameter Definitions
"""

from meteostat.enumerations import Granularity, Parameter, Unit
from meteostat.typing import ParameterSpec
from meteostat.utils.validators import maximum, minimum


DEFAULT_PARAMETERS = [
    ParameterSpec(
        id=Parameter.TEMP,
        name="Air Temperature",
        granularity=Granularity.HOURLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TEMP,
        name="Mean Air Temperature",
        granularity=Granularity.DAILY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TEMP,
        name="Mean Air Temperature",
        granularity=Granularity.MONTHLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TEMP,
        name="Mean Air Temperature",
        granularity=Granularity.NORMALS,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMIN,
        name="Minimum Air Temperature",
        granularity=Granularity.DAILY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMIN,
        name="Mean Daily Minimum Air Temperature",
        granularity=Granularity.MONTHLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMIN,
        name="Mean Daily Minimum Air Temperature",
        granularity=Granularity.NORMALS,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMAX,
        name="Maximum Air Temperature",
        granularity=Granularity.DAILY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMAX,
        name="Mean Daily Maximum Air Temperature",
        granularity=Granularity.MONTHLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TMAX,
        name="Mean Daily Maximum Air Temperature",
        granularity=Granularity.NORMALS,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TXMN,
        name="Absolute Minimum Air Temperature",
        granularity=Granularity.MONTHLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TXMN,
        name="Absolute Minimum Air Temperature",
        granularity=Granularity.NORMALS,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TXMX,
        name="Absolute Maximum Air Temperature",
        granularity=Granularity.MONTHLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.TXMX,
        name="Absolute Maximum Air Temperature",
        granularity=Granularity.NORMALS,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.DWPT,
        name="Dew Point",
        granularity=Granularity.HOURLY,
        unit=Unit.CELSIUS,
        dtype="Float64",
        validators=[minimum(-100), maximum(65)],
    ),
    ParameterSpec(
        id=Parameter.RHUM,
        name="Relative Humidity",
        granularity=Granularity.HOURLY,
        unit=Unit.PERCENTAGE,
        dtype="UInt8",
        validators=[minimum(0), maximum(100)],
    ),
    ParameterSpec(
        id=Parameter.RHUM,
        name="Relative Humidity",
        granularity=Granularity.DAILY,
        unit=Unit.PERCENTAGE,
        dtype="UInt8",
        validators=[minimum(0), maximum(100)],
    ),
    ParameterSpec(
        id=Parameter.RHUM,
        name="Relative Humidity",
        granularity=Granularity.MONTHLY,
        unit=Unit.PERCENTAGE,
        dtype="UInt8",
        validators=[minimum(0), maximum(100)],
    ),
    ParameterSpec(
        id=Parameter.RHUM,
        name="Relative Humidity",
        granularity=Granularity.NORMALS,
        unit=Unit.PERCENTAGE,
        dtype="UInt8",
        validators=[minimum(0), maximum(100)],
    ),
    ParameterSpec(
        id=Parameter.PRCP,
        name="Total Precipitation",
        granularity=Granularity.HOURLY,
        unit=Unit.MILLIMETERS,
        dtype="Float64",
        validators=[minimum(0), maximum(350)],
    ),
    ParameterSpec(
        id=Parameter.PRCP,
        name="Total Precipitation",
        granularity=Granularity.DAILY,
        unit=Unit.MILLIMETERS,
        dtype="Float64",
        validators=[minimum(0), maximum(2000)],
    ),
    ParameterSpec(
        id=Parameter.PRCP,
        name="Total Precipitation",
        granularity=Granularity.MONTHLY,
        unit=Unit.MILLIMETERS,
        dtype="Float64",
        validators=[minimum(0), maximum(10000)],
    ),
    ParameterSpec(
        id=Parameter.PRCP,
        name="Total Precipitation",
        granularity=Granularity.NORMALS,
        unit=Unit.MILLIMETERS,
        dtype="Float64",
        validators=[minimum(0), maximum(10000)],
    ),
    ParameterSpec(
        id=Parameter.SNOW,
        name="Snowfall",
        granularity=Granularity.HOURLY,
        unit=Unit.CENTIMETERS,
        dtype="UInt8",
        validators=[minimum(0), maximum(10000)],
    ),
    ParameterSpec(
        id=Parameter.SNWD,
        name="Snow Depth",
        granularity=Granularity.HOURLY,
        unit=Unit.CENTIMETERS,
        dtype="UInt16",
        validators=[minimum(0), maximum(1200)],
    ),
    ParameterSpec(
        id=Parameter.SNWD,
        name="Snow Depth",
        granularity=Granularity.DAILY,
        unit=Unit.CENTIMETERS,
        dtype="UInt16",
        validators=[minimum(0), maximum(1200)],
    ),
    ParameterSpec(
        id=Parameter.WDIR,
        name="Wind Direction",
        granularity=Granularity.HOURLY,
        unit=Unit.DEGREES,
        dtype="UInt16",
        validators=[minimum(0), maximum(360)],
    ),
    ParameterSpec(
        id=Parameter.WSPD,
        name="Wind Speed",
        granularity=Granularity.HOURLY,
        unit=Unit.KMH,
        dtype="Float64",
        validators=[minimum(0), maximum(250)],
    ),
    ParameterSpec(
        id=Parameter.WSPD,
        name="Average Wind Speed",
        granularity=Granularity.DAILY,
        unit=Unit.KMH,
        dtype="Float64",
        validators=[minimum(0), maximum(150)],
    ),
    ParameterSpec(
        id=Parameter.WPGT,
        name="Peak Wind Gust",
        granularity=Granularity.HOURLY,
        unit=Unit.KMH,
        dtype="Float64",
        validators=[minimum(0), maximum(500)],
    ),
    ParameterSpec(
        id=Parameter.WPGT,
        name="Peak Wind Gust",
        granularity=Granularity.DAILY,
        unit=Unit.KMH,
        dtype="Float64",
        validators=[minimum(0), maximum(500)],
    ),
    ParameterSpec(
        id=Parameter.PRES,
        name="Air Pressure (MSL)",
        granularity=Granularity.HOURLY,
        unit=Unit.HPA,
        dtype="Float64",
        validators=[minimum(850), maximum(1090)],
    ),
    ParameterSpec(
        id=Parameter.PRES,
        name="Average Air Pressure (MSL)",
        granularity=Granularity.DAILY,
        unit=Unit.HPA,
        dtype="Float64",
        validators=[minimum(850), maximum(1090)],
    ),
    ParameterSpec(
        id=Parameter.PRES,
        name="Average Air Pressure (MSL)",
        granularity=Granularity.MONTHLY,
        unit=Unit.HPA,
        dtype="Float64",
        validators=[minimum(850), maximum(1090)],
    ),
    ParameterSpec(
        id=Parameter.PRES,
        name="Average Air Pressure (MSL)",
        granularity=Granularity.NORMALS,
        unit=Unit.HPA,
        dtype="Float64",
        validators=[minimum(850), maximum(1090)],
    ),
    ParameterSpec(
        id=Parameter.TSUN,
        name="Sunshine Duration",
        granularity=Granularity.HOURLY,
        unit=Unit.MINUTES,
        dtype="UInt8",
        validators=[minimum(0), maximum(60)],
    ),
    ParameterSpec(
        id=Parameter.TSUN,
        name="Sunshine Duration",
        granularity=Granularity.DAILY,
        unit=Unit.MINUTES,
        dtype="UInt16",
        validators=[minimum(0), maximum(1440)],
    ),
    ParameterSpec(
        id=Parameter.TSUN,
        name="Sunshine Duration",
        granularity=Granularity.MONTHLY,
        unit=Unit.MINUTES,
        dtype="UInt16",
        validators=[minimum(0), maximum(44640)],
    ),
    ParameterSpec(
        id=Parameter.TSUN,
        name="Sunshine Duration",
        granularity=Granularity.NORMALS,
        unit=Unit.MINUTES,
        dtype="UInt16",
        validators=[minimum(0), maximum(44640)],
    ),
    ParameterSpec(
        id=Parameter.CLDC,
        name="Cloud Cover",
        granularity=Granularity.HOURLY,
        unit=Unit.OKTAS,
        dtype="UInt8",
        validators=[minimum(0), maximum(8)],
    ),
    ParameterSpec(
        id=Parameter.CLDC,
        name="Average Cloud Cover",
        granularity=Granularity.DAILY,
        unit=Unit.OKTAS,
        dtype="UInt8",
        validators=[minimum(0), maximum(8)],
    ),
    ParameterSpec(
        id=Parameter.VSBY,
        name="Visibility",
        granularity=Granularity.HOURLY,
        unit=Unit.METERS,
        dtype="UInt16",
        validators=[minimum(0)],
    ),
    ParameterSpec(
        id=Parameter.COCO,
        name="Weather Condition Code",
        granularity=Granularity.HOURLY,
        dtype="UInt8",
        validators=[minimum(0), maximum(27)],
    ),
]



"""
Climate Normals

Access climate normals data for one or multiple weather stations.
"""

from typing import List, Optional

import numpy as np
import pandas as pd

from meteostat.enumerations import Parameter, Provider, Granularity
from meteostat.core.schema import schema_service
from meteostat.api.monthly import DEFAULT_PARAMETERS, monthly
from meteostat.api.timeseries import TimeSeries
from meteostat.typing import Station
from meteostat.api.point import Point
from meteostat.utils.data import reshape_by_source
from meteostat.utils.parsers import parse_year


def normals(
    station: str | Station | Point | List[str | Station | Point] | pd.DataFrame,
    start: int = 1961,
    end: int = 1990,
    parameters: Optional[List[Parameter]] = None,
    providers: Optional[List[Provider]] = None,
    max_missing: int = 3,
):
    """
    Access climate normals data.

    Parameters
    ----------
    station : str, Station, Point, List[str | Station | Point], pd.Index, pd.Series
        Weather station(s) or Point(s) to query data for. Can be a single station/point or a list.
        Points are converted to virtual stations with IDs like $0001, $0002, etc.
    start : int, optional
        Start year for the data query. Defaults to 1961.
    end : int, optional
        End year for the data query. Defaults to 1990.
    parameters : List[Parameter], optional
        List of parameters to include in the data query. Defaults to a set of common parameters.
    providers : List[Provider], optional
        List of data providers to use for the query. Defaults to the monthly provider.
    max_missing : int, optional
        Maximum number of missing values allowed in a month to calculate the mean. Defaults to 3.

    Returns
    -------
    TimeSeries
        A TimeSeries object containing the climate normals data for the specified
        stations and parameters.
    """
    if parameters is None:
        parameters = DEFAULT_PARAMETERS
    if providers is None:
        providers = [Provider.MONTHLY]

    def _mean(group: pd.Series):
        """
        Calculate the monthly mean from multiple years of monthly data
        """
        if group.isna().sum() > max_missing:
            return np.nan
        return group.mean()

    # Fetch monthly data for the specified stations and parameters
    ts = monthly(
        station,
        parse_year(start),
        parse_year(end, True),
        parameters=parameters,
        providers=providers,
    )
    df = ts.fetch(sources=True)

    # Process data if available
    if df is not None and not df.empty:
        # Add station level if only a single station is provided
        if not ts._multi_station:
            df = pd.concat(
                [df],
                keys=[ts.stations.index.get_level_values("id")[0]],
                names=["station"],
            )

        # Extract month from time index
        df["month"] = df.index.get_level_values("time").month

        # Create aggregation functions for different column types
        agg_funcs = {}
        for col in df.columns:
            if col.endswith("_source"):
                # For source columns, concatenate unique values
                agg_funcs[col] = lambda x: ", ".join(x.dropna().unique())
            elif col == "txmx":
                # For temperature maximum, use max function
                agg_funcs[col] = lambda x: x.max() if not x.isna().all() else np.nan
            elif col == "txmn":
                # For temperature minimum, use min function
                agg_funcs[col] = lambda x: x.min() if not x.isna().all() else np.nan
            else:
                # For data columns, use the mean function
                agg_funcs[col] = _mean

        # Apply aggregation functions
        df = df.groupby(["station", "month"]).agg(agg_funcs)
        df = df.rename_axis(index={"month": "time"})

        # Separate parameter columns from source columns for formatting
        param_cols = [str(param) for param in parameters if str(param) in df.columns]
        source_cols = [col for col in df.columns if col.endswith("_source")]

        # Format only the parameter columns
        if param_cols:
            df_data = schema_service.format(df[param_cols], Granularity.NORMALS)
            # Combine formatted parameter columns with unformatted source columns
            if source_cols:
                df = pd.concat([df_data, df[source_cols]], axis=1)
            else:
                df = df_data

        # Reshape data by source for each station
        df_fragments = []
        for s in df.index.get_level_values("station").unique():
            station_df = df.loc[s]
            fragment = reshape_by_source(station_df)
            fragment = pd.concat([fragment], keys=[s], names=["station"])
            df_fragments.append(fragment)

        df = pd.concat(df_fragments)
    else:
        df = None

    # Return the final TimeSeries object
    return TimeSeries(
        granularity=Granularity.NORMALS,
        stations=ts.stations,
        df=df,
        start=ts.start,
        end=ts.end,
        multi_station=ts._multi_station,
    )



from datetime import datetime
from typing import Optional, Union
from urllib.error import HTTPError

from numpy import isnan
import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.core.logger import logger
from meteostat.core.cache import cache_service
from meteostat.typing import ProviderRequest
from meteostat.utils.conversions import ms_to_kmh, temp_dwpt_to_rhum

ISD_LITE_ENDPOINT = "https://www.ncei.noaa.gov/pub/data/noaa/isd-lite/"
COLSPECS = [
    (0, 4),
    (5, 7),
    (8, 10),
    (11, 13),
    (13, 19),
    (19, 25),
    (25, 31),
    (31, 37),
    (37, 43),
    (43, 49),
    (49, 55),
]
COLUMN_NAMES = [
    "time",
    Parameter.TEMP,
    Parameter.DWPT,
    Parameter.PRES,
    Parameter.WDIR,
    Parameter.WSPD,
    Parameter.CLDC,
    Parameter.PRCP,
]


def map_sky_code(code: Union[int, str]) -> Optional[int]:
    """
    Only accept okta
    """
    return int(code) if not isnan(code) and int(code) >= 0 and int(code) <= 8 else None


def get_ttl(_usaf: str, _wban: str, year: int) -> int:
    """
    Get TTL based on year

    Current + previous year = one day
    Else = 30 days
    """
    current_year = datetime.now().year
    return TTL.DAY if current_year - year < 2 else TTL.MONTH


@cache_service.cache(get_ttl, "pickle")
def get_df(usaf: str, wban: str, year: int) -> Optional[pd.DataFrame]:
    if not usaf:
        return None

    filename = f"{usaf}-{wban if wban else '99999'}-{year}.gz"

    try:
        df = pd.read_fwf(
            f"{ISD_LITE_ENDPOINT}{year}/{filename}",
            na_values=["-9999", -9999],
            header=None,
            colspecs=COLSPECS,
            compression="gzip",
        )

        # Parse datetime from first 4 columns (year, month, day, hour)
        df[0] = pd.to_datetime(
            df[0].astype(str)
            + "-"
            + df[1].astype(str).str.zfill(2)
            + "-"
            + df[2].astype(str).str.zfill(2)
            + " "
            + df[3].astype(str).str.zfill(2)
            + ":00",
            format="%Y-%m-%d %H:%M",
        )
        df = df.drop(columns=[1, 2, 3])

        # Rename columns
        df.columns = COLUMN_NAMES

        # Adapt columns
        df[Parameter.TEMP] = df[Parameter.TEMP].div(10)
        df[Parameter.DWPT] = df[Parameter.DWPT].div(10)
        df[Parameter.PRES] = df[Parameter.PRES].div(10)
        df[Parameter.WSPD] = df[Parameter.WSPD].div(10).apply(ms_to_kmh)
        df[Parameter.CLDC] = df[Parameter.CLDC].apply(map_sky_code)
        df[Parameter.PRCP] = df[Parameter.PRCP].div(10)

        # Calculate humidity data
        # pylint: disable=unnecessary-lambda
        df[Parameter.RHUM] = df.apply(lambda row: temp_dwpt_to_rhum(row), axis=1)

        # Drop dew point column
        # pylint: disable=no-member
        df = df.drop(Parameter.DWPT, axis=1)

        # Set index
        df = df.set_index("time")

        # Round decimals
        return df.round(1)

    except HTTPError as error:
        if error.status == 404:
            logger.info(f"ISD Lite file not found: {filename}")
        else:
            logger.warning(
                f"Couldn't load ISD Lite file {filename} (status: {error.status})"
            )
        return None

    except Exception as error:
        logger.warning(error)
        return None


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """ """
    if req.start is None or req.end is None:
        return None

    years = range(req.start.year, req.end.year + 1)
    data = tuple(
        map(
            lambda i: get_df(*i),
            (
                (
                    (
                        req.station.identifiers["usaf"]
                        if "usaf" in req.station.identifiers
                        else None
                    ),
                    (
                        req.station.identifiers["wban"]
                        if "wban" in req.station.identifiers
                        else None
                    ),
                    year,
                )
                for year in years
            ),
        )
    )

    return pd.concat(data) if len(data) and not all(d is None for d in data) else None



"""
Parameter Service

The Parameter Service provides methods to manage and access
supported parameters for data requests.
"""

from typing import List, Optional

from meteostat.core.logger import logger
from meteostat.enumerations import Granularity, Parameter
from meteostat.parameters import DEFAULT_PARAMETERS
from meteostat.typing import ParameterSpec


class ParameterService:
    """
    Parameter Service
    """

    _parameters: List[ParameterSpec]

    @staticmethod
    def _has_duplicates(parameter_specs: List[ParameterSpec]) -> bool:
        """
        Check if parameter list contains duplicates
        """
        seen = set()
        for spec in parameter_specs:
            key = (spec.id, spec.granularity)
            if key in seen:
                return True  # Duplicate found
            seen.add(key)
        return False  # No duplicates found

    def _parameter_exists(self, parameter: ParameterSpec) -> bool:
        """
        Check if a parameter already exists
        """
        key = (parameter.id, parameter.granularity)
        return any((spec.id, spec.granularity) == key for spec in self.parameters)

    def __init__(self, parameters: List[ParameterSpec]) -> None:
        if self._has_duplicates(parameters):
            raise ValueError("List of parameters contains duplicates")

        self._parameters = parameters

    @property
    def parameters(self) -> List[ParameterSpec]:
        """
        Get supported parameters
        """
        return self._parameters

    def register(self, parameter: ParameterSpec) -> None:
        """
        Register a parameter
        """
        if self._parameter_exists(parameter):
            raise ValueError("The parameter already exists")

        self._parameters.append(parameter)

    def get_parameter(
        self, parameter_id: Parameter, granularity: Granularity
    ) -> Optional[ParameterSpec]:
        """
        Get parameter by ID and granularity
        """
        return next(
            (
                parameter
                for parameter in self.parameters
                if parameter.id == parameter_id and parameter.granularity == granularity
            ),
            None,
        )

    def filter_parameters(
        self, granularity: Granularity, parameters: List[Parameter]
    ) -> List[Parameter]:
        """
        Raise exception if a requested parameter is not part of the schema
        """
        supported_parameters = list(
            map(
                lambda parameter: parameter.id,
                filter(
                    lambda parameter: parameter.granularity == granularity,
                    self.parameters,
                ),
            )
        )
        # Get difference between requested parameters and root schema
        diff = set(parameters).difference(supported_parameters)
        # Log warning
        if diff:
            logger.error(
                "Tried to request data for unsupported parameter(s): %s",
                ", ".join(diff),
            )
        # Return intersection
        return list(
            filter(
                lambda parameter: parameter in parameters,
                supported_parameters,
            )
        )


parameter_service = ParameterService(DEFAULT_PARAMETERS)



"""
Interpolation Module

Provides spatial interpolation functions for meteorological data.
"""

from typing import Optional, Union

import numpy as np
import pandas as pd

from meteostat.api.point import Point
from meteostat.api.timeseries import TimeSeries
from meteostat.typing import Station
from meteostat.enumerations import Parameter
from meteostat.interpolation.lapserate import apply_lapse_rate
from meteostat.interpolation.nearest import nearest_neighbor
from meteostat.interpolation.idw import inverse_distance_weighting
from meteostat.utils.data import aggregate_sources, reshape_by_source, stations_to_df
from meteostat.utils.geo import get_distance
from meteostat.utils.parsers import parse_station
from meteostat.core.schema import schema_service
from meteostat.core.logger import logger


# Parameters that are categorical and should not use IDW interpolation
CATEGORICAL_PARAMETERS = {Parameter.WDIR, Parameter.CLDC, Parameter.COCO}


def _create_timeseries(
    ts: TimeSeries, point: Point, df: Optional[pd.DataFrame] = None
) -> TimeSeries:
    """
    Create a TimeSeries object from interpolated DataFrame
    """
    parsed = parse_station(point)
    stations_list = [parsed] if isinstance(parsed, Station) else parsed

    # Convert stations to DataFrame
    stations_df = stations_to_df(stations_list)

    return TimeSeries(
        ts.granularity,
        stations_df,
        df=df,
        start=ts.start,
        end=ts.end,
        timezone=ts.timezone,
    )


def _add_source_columns(
    result: pd.DataFrame,
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Add source columns to the result DataFrame
    """
    source_cols = [c for c in df.columns if c.endswith("_source")]
    if source_cols:
        grouped = df.groupby("time")[source_cols].agg(aggregate_sources)
        if isinstance(grouped, pd.Series):
            grouped = grouped.to_frame(name=source_cols[0])
        grouped.index.name = "time"

        # Safely align on time and add/fill source columns without causing overlaps
        result_has_time_col = "time" in result.columns
        if result_has_time_col:
            result = result.set_index("time")

        # Ensure both frames align on the same index (time)
        # For each source column, add it if missing or fill NaNs if present
        for col in source_cols:
            if col in grouped.columns:
                if col in result.columns:
                    # Fill missing values in result using aggregated sources
                    result[col] = result[col].where(result[col].notna(), grouped[col])
                else:
                    # Add aggregated source column
                    result[col] = grouped[col]

        if result_has_time_col:
            result = result.reset_index()
    return result


def _prepare_data_with_distances(
    df: pd.DataFrame, point: Point, elevation_weight: float
) -> pd.DataFrame:
    """
    Add distance and elevation calculations to the DataFrame
    """
    # Add distance column
    df["distance"] = get_distance(
        point.latitude, point.longitude, df["latitude"], df["longitude"]
    )

    # Add effective distance column if elevation is available
    if point.elevation is not None and "elevation" in df.columns:
        elev_diff = np.abs(df["elevation"] - point.elevation)
        df["effective_distance"] = np.sqrt(
            df["distance"] ** 2 + (elev_diff * elevation_weight) ** 2
        )
    else:
        df["effective_distance"] = df["distance"]

    # Add elevation difference column
    if "elevation" in df.columns and point.elevation is not None:
        df["elevation_diff"] = np.abs(df["elevation"] - point.elevation)
    else:
        df["elevation_diff"] = np.nan

    return df


def _should_use_nearest_neighbor(
    df: pd.DataFrame,
    point: Point,
    distance_threshold: Union[int, None],
    elevation_threshold: Union[int, None],
) -> bool:
    """
    Determine if nearest neighbor should be used based on thresholds
    """
    min_distance = df["distance"].min()
    use_nearest = distance_threshold is None or min_distance <= distance_threshold

    if use_nearest and point.elevation is not None and "elevation" in df.columns:
        min_elev_diff = np.abs(df["elevation"] - point.elevation).min()
        use_nearest = (
            elevation_threshold is None or min_elev_diff <= elevation_threshold
        )

    return use_nearest


def _get_categorical_columns(df: pd.DataFrame) -> list:
    """
    Identify categorical columns in the data (excluding source columns)
    """
    data_cols = [c for c in df.columns if not c.endswith("_source")]
    return [c for c in data_cols if c in CATEGORICAL_PARAMETERS]


def _interpolate_with_nearest_neighbor(
    df: pd.DataFrame,
    ts: TimeSeries,
    point: Point,
    distance_threshold: Union[int, None],
    elevation_threshold: Union[int, None],
) -> Optional[pd.DataFrame]:
    """
    Perform nearest neighbor interpolation with threshold filtering
    """
    distance_filter = (
        pd.Series([True] * len(df), index=df.index)
        if distance_threshold is None
        else (df["distance"] <= distance_threshold)
    )
    elevation_filter = (
        pd.Series([True] * len(df), index=df.index)
        if elevation_threshold is None
        else (np.abs(df["elevation"] - point.elevation) <= elevation_threshold)
    )
    df_filtered = df[distance_filter & elevation_filter]
    return nearest_neighbor(df_filtered, ts, point)


def _interpolate_with_idw_and_categorical(
    df: pd.DataFrame,
    ts: TimeSeries,
    point: Point,
    categorical_cols: list,
    power: float,
) -> Optional[pd.DataFrame]:
    """
    Perform IDW interpolation for non-categorical parameters and nearest neighbor for categorical
    """
    # For categorical parameters, always use nearest neighbor
    if categorical_cols:
        df_categorical = nearest_neighbor(df, ts, point)
        # Keep only categorical columns that exist in the result
        existing_categorical = [
            c for c in categorical_cols if c in df_categorical.columns
        ]
        df_categorical = (
            df_categorical[existing_categorical]
            if existing_categorical
            else pd.DataFrame()
        )
    else:
        df_categorical = pd.DataFrame()

    # Perform IDW interpolation for all parameters
    idw_func = inverse_distance_weighting(power=power)
    df_idw = idw_func(df, ts, point)

    # Remove categorical columns from IDW result if they exist
    if not df_categorical.empty and df_idw is not None:
        # Drop categorical columns from IDW result
        idw_cols_to_keep = [c for c in df_idw.columns if c not in categorical_cols]
        df_idw = df_idw[idw_cols_to_keep] if idw_cols_to_keep else pd.DataFrame()

    # Combine categorical (nearest) and non-categorical (IDW) results
    if not df_categorical.empty and not df_idw.empty:
        return pd.concat([df_idw, df_categorical], axis=1)
    elif not df_categorical.empty:
        return df_categorical
    else:
        return df_idw


def _merge_interpolation_results(
    df_nearest: Optional[pd.DataFrame],
    df_idw: Optional[pd.DataFrame],
    use_nearest: bool,
) -> Optional[pd.DataFrame]:
    """
    Merge nearest neighbor and IDW results with appropriate priority
    """
    if use_nearest and df_nearest is not None and len(df_nearest) > 0:
        if df_idw is not None:
            # Combine nearest and IDW results, prioritizing nearest values
            return df_nearest.combine_first(df_idw)
        else:
            return df_nearest
    else:
        return df_idw


def _postprocess_result(
    result: pd.DataFrame, df: pd.DataFrame, ts: TimeSeries
) -> pd.DataFrame:
    """
    Post-process the interpolation result: drop location columns, add sources, format, reshape
    """
    # Drop location-related columns
    result = result.drop(
        [
            "latitude",
            "longitude",
            "elevation",
            "distance",
            "effective_distance",
            "elevation_diff",
        ],
        axis=1,
        errors="ignore",
    )

    # Add source columns
    result = _add_source_columns(result, df)

    # Reshape by source
    result = reshape_by_source(result)

    # Add station index
    result["station"] = "$0001"
    result = result.set_index("station", append=True).reorder_levels(
        ["station", "time", "source"]
    )

    # Reorder columns to match the canonical schema order
    result = schema_service.purge(result, ts.parameters)

    # Format the result using schema_service to apply proper rounding
    result = schema_service.format(result, ts.granularity)

    return result


def interpolate(
    ts: TimeSeries,
    point: Point,
    distance_threshold: Union[int, None] = 5000,
    elevation_threshold: Union[int, None] = 50,
    elevation_weight: float = 10,
    power: float = 2.0,
    lapse_rate: Union[float, None] = 6.5,
    lapse_rate_threshold: int = 50,
) -> TimeSeries:
    """
    Interpolate time series data spatially to a specific point.

    Parameters
    ----------
    ts : TimeSeries
        The time series to interpolate.
    point : Point
        The point to interpolate the data for.
    distance_threshold : int, optional
        Maximum distance (in meters) to use nearest neighbor (default: 5000).
        Beyond this, IDW is used.
    elevation_threshold : int, optional
        Maximum elevation difference (in meters) to use nearest neighbor (default: 50).
        Beyond this, IDW is used even if distance is within threshold.
    elevation_weight : float, optional
        Weight for elevation difference in distance calculation (default: 0.1).
        The effective distance is calculated as:
        sqrt(horizontal_distance^2 + (elevation_diff * elevation_weight)^2)
    power : float, optional
        Power parameter for IDW (default: 2.0). Higher values give more
        weight to closer stations.
    lapse_rate : float, optional
        Apply lapse rate correction based on elevation difference (default: 6.5).
    lapse_rate_threshold : int, optional
        Elevation difference threshold (in meters) to apply lapse rate correction
        (default: 50). If the elevation difference between the point and stations
        is less than this, no correction is applied.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame containing the interpolated data for the specified point,
        or None if no data is available.
    """
    # Fetch DataFrame, filling missing values and adding location data
    df = ts.fetch(fill=True, location=True, sources=True)

    # If no data is returned, return None
    if df is None:
        logger.debug("No data available for interpolation. Returning empty TimeSeries.")
        return _create_timeseries(ts, point)

    # Prepare data with distance and elevation calculations
    df = _prepare_data_with_distances(df, point, elevation_weight)

    # Apply lapse rate if specified and elevation is available
    if (
        lapse_rate
        and point.elevation
        and df["elevation_diff"].max() >= lapse_rate_threshold
    ):
        logger.debug("Applying lapse rate correction.")
        df = apply_lapse_rate(df, point.elevation, lapse_rate)

    # Determine if nearest neighbor should be used
    use_nearest = _should_use_nearest_neighbor(
        df, point, distance_threshold, elevation_threshold
    )

    # Identify categorical columns
    categorical_cols = _get_categorical_columns(df)
    logger.debug(f"Categorical columns identified: {categorical_cols}")

    # Perform interpolation
    df_nearest = None
    df_idw = None

    if use_nearest:
        logger.debug("Using nearest neighbor interpolation.")
        df_nearest = _interpolate_with_nearest_neighbor(
            df, ts, point, distance_threshold, elevation_threshold
        )

    # Use IDW if nearest neighbor doesn't provide complete data
    if (
        not use_nearest
        or df_nearest is None
        or len(df_nearest) == 0
        or df_nearest.isna().any().any()
    ):
        logger.debug("Using IDW interpolation.")
        df_idw = _interpolate_with_idw_and_categorical(
            df, ts, point, categorical_cols, power
        )

    # Merge results
    result = _merge_interpolation_results(df_nearest, df_idw, use_nearest)

    # If no data is returned, return None
    if result is None or result.empty:
        return _create_timeseries(ts, point)

    # Post-process result
    result = _postprocess_result(result, df, ts)

    return _create_timeseries(ts, point, result)



"""
Type parsing utilities

This module contains utilities for extracting and validating types
from class annotations without creating circular imports.
"""

import types
from typing import Any, Union, get_origin


def _is_union_type(annotation: Any) -> bool:
    """
    Check if an annotation is a Union type (either typing.Union or types.UnionType).

    Supports both Optional[X] (typing.Union[X, None]) and X | None (types.UnionType).
    """
    # Python 3.10+ union syntax (X | Y)
    if isinstance(annotation, types.UnionType):
        return True
    # typing.Union syntax
    return hasattr(annotation, "__origin__") and annotation.__origin__ is Union


def _get_union_args(annotation: Any) -> tuple:
    """
    Get the arguments of a Union type annotation.

    Works with both typing.Union and types.UnionType.
    """
    if isinstance(annotation, types.UnionType):
        return annotation.__args__
    return getattr(annotation, "__args__", ())


def extract_property_type(cls: type, property_name: str) -> tuple[Any, Any]:
    """
    Extract the expected type for a class property from type annotations.

    Parameters
    ----------
    cls : type
        The class to extract type information from
    property_name : str
        The name of the property to get the type for

    Returns
    -------
    tuple[Any, Any]
        A tuple containing (expected_type, original_type) where:
        - expected_type: The actual type to validate against (unwrapped from Optional)
        - original_type: The original type annotation (including Optional wrapper)

    Raises
    ------
    ValueError
        If the property doesn't exist on the class
    """
    if not hasattr(cls, property_name):
        raise ValueError(f"Property '{property_name}' does not exist")

    # Get type annotations for the class
    annotations = getattr(cls, "__annotations__", {})
    original_type = annotations.get(property_name)

    if original_type is None:
        return None, None

    expected_type = original_type

    # Handle Optional types (extract the inner type)
    # Supports both Optional[X] and X | None syntax
    if _is_union_type(original_type):
        args = _get_union_args(original_type)
        if len(args) == 2 and type(None) in args:
            # This is Optional[Type], extract the non-None type
            # Type narrowing: we know len(args) == 2, so both indices are valid
            arg0 = args[0]  # type: ignore
            arg1 = args[1]  # type: ignore
            expected_type = arg0 if arg1 is type(None) else arg1

    return expected_type, original_type


def validate_parsed_value(
    value: Any, expected_type: Any, original_type: Any, property_name: str
) -> Any:
    """
    Validate a parsed value against the expected type and handle special cases.

    Parameters
    ----------
    value : Any
        The parsed value to validate
    expected_type : Any
        The expected type for validation
    original_type : Any
        The original type annotation (for Optional type checking)
    property_name : str
        The name of the property being validated (for error messages)

    Returns
    -------
    Any
        The validated (and potentially converted) value

    Raises
    ------
    ValueError
        If the value doesn't match the expected type
    TypeError
        If the expected type cannot be used for isinstance checks
    """
    # Extract type name once for error messages
    type_name = getattr(expected_type, "__name__", str(expected_type))

    # Special case for Optional types - None is allowed
    # Supports both Optional[X] and X | None syntax
    if _is_union_type(original_type) and value is None:
        args = _get_union_args(original_type)
        if len(args) == 2 and type(None) in args:
            return value

    # Special case for bool type - allow 0 and 1 to be parsed as False and True
    if expected_type is bool and isinstance(value, int):
        if value == 0:
            return False
        if value == 1:
            return True
        raise ValueError(
            f"Environment variable '{property_name}' has value {value} "
            f"but boolean type only accepts 0, 1, true, or false"
        )

    # Get the origin type for parameterized generics (e.g., list[str] -> list)
    origin_type = get_origin(expected_type)

    # If it's a parameterized generic, validate against the origin type
    if origin_type is not None:
        # For parameterized generics like list[str], validate against the origin (list)
        if isinstance(value, origin_type):
            return value

        # Type mismatch for parameterized generic
        raise ValueError(
            f"Environment variable '{property_name}' has type {type(value).__name__} "
            f"but expected {type_name}"
        )

    # For non-generic types, use isinstance directly
    try:
        if isinstance(value, expected_type):
            return value
    except TypeError as e:
        # isinstance can fail for some types, convert to ValueError with context
        raise TypeError(
            f"Cannot validate type for environment variable '{property_name}': {e}"
        ) from e

    # Type mismatch
    raise ValueError(
        f"Environment variable '{property_name}' has type {type(value).__name__} "
        f"but expected {type_name}"
    )



from typing import Optional

import numpy as np
import pandas as pd

from meteostat.enumerations import Parameter, Provider
from meteostat.api.daily import daily
from meteostat.typing import ProviderRequest
from meteostat.utils.data import aggregate_sources, reshape_by_source
from meteostat.utils.parsers import parse_month


def monthly_mean(group: pd.Series):
    """
    Calculate the monthly mean from a series of daily data
    """
    if group.isna().sum() > 3:
        return np.nan
    return group.interpolate(axis=0).mean()


def monthly_sum(group: pd.Series):
    """
    Calculate the monthly sum from a series of daily data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.sum()


def monthly_min(group: pd.Series):
    """
    Calculate the absolute minimum from a series of daily data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.interpolate(axis=0).min()


def monthly_max(group: pd.Series):
    """
    Calculate the absolute maximum from a series of daily data
    """
    if group.isna().sum() > 0:
        return np.nan
    return group.interpolate(axis=0).max()


# Available parameters with source column and aggregation method
PARAMETER_AGGS = {
    Parameter.TEMP: (Parameter.TEMP, monthly_mean),
    Parameter.TMIN: (Parameter.TMIN, monthly_mean),
    Parameter.TMAX: (Parameter.TMAX, monthly_mean),
    Parameter.TXMN: (Parameter.TMIN, monthly_min),
    Parameter.TXMX: (Parameter.TMAX, monthly_max),
    Parameter.PRCP: (Parameter.PRCP, monthly_sum),
    Parameter.PRES: (Parameter.PRES, monthly_mean),
    Parameter.TSUN: (Parameter.TSUN, monthly_sum),
}


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """
    Fetch daily weather data from Meteostat's central data
    repository and aggregate to monthly granularity
    """
    # Get all source columns
    source_cols = list(dict.fromkeys([PARAMETER_AGGS[p][0] for p in req.parameters]))

    # Get daily DataFrame
    ts_daily = daily(
        req.station.id,
        parse_month(req.start),
        parse_month(req.end, is_end=True),
        parameters=source_cols,
        providers=[Provider.DAILY],
    )

    df_daily = ts_daily.fetch(fill=True, sources=True)

    # If no daily data is available, exit
    if df_daily is None:
        return None

    # Create monthly aggregations
    df = pd.DataFrame()
    for parameter in req.parameters:
        [daily_param_name, agg_func] = PARAMETER_AGGS[parameter]
        df[parameter] = (
            df_daily[daily_param_name]
            .groupby(pd.Grouper(level="time", freq="MS"))
            .agg(agg_func)
        )
        df[f"{parameter}_source"] = (
            df_daily[f"{daily_param_name}_source"]
            .groupby(pd.Grouper(level="time", freq="MS"))
            .agg(aggregate_sources)
        )

    # Adjust DataFrame and add index
    df = df.round(1)
    df.index = pd.to_datetime(df.index.astype("datetime64").dt.date)  # type: ignore[union-attr]
    df.index.name = "time"

    # Return final DataFrame
    return reshape_by_source(df)



from typing import Optional, Union
from urllib.error import HTTPError

import pandas as pd

from meteostat.api.config import config
from meteostat.enumerations import TTL, Parameter
from meteostat.core.logger import logger
from meteostat.core.network import network_service
from meteostat.typing import ProviderRequest
from meteostat.utils.conversions import percentage_to_okta
from meteostat.core.cache import cache_service

CONDICODES = {
    "clearsky": 1,
    "cloudy": 3,
    "fair": 2,
    "fog": 5,
    "heavyrain": 9,
    "heavyrainandthunder": 26,
    "heavyrainshowers": 18,
    "heavyrainshowersandthunder": 26,
    "heavysleet": 13,
    "heavysleetandthunder": 26,
    "heavysleetshowers": 20,
    "heavysleetshowersandthunder": 26,
    "heavysnow": 16,
    "heavysnowandthunder": 26,
    "heavysnowshowers": 22,
    "heavysnowshowersandthunder": 26,
    "lightrain": 7,
    "lightrainandthunder": 25,
    "lightrainshowers": 17,
    "lightrainshowersandthunder": 25,
    "lightsleet": 12,
    "lightsleetandthunder": 25,
    "lightsleetshowers": 19,
    "lightsnow": 14,
    "lightsnowandthunder": 25,
    "lightsnowshowers": 21,
    "lightssleetshowersandthunder": 25,
    "lightssnowshowersandthunder": 25,
    "partlycloudy": 3,
    "rain": 8,
    "rainandthunder": 25,
    "rainshowers": 17,
    "rainshowersandthunder": 25,
    "sleet": 12,
    "sleetandthunder": 25,
    "sleetshowers": 19,
    "sleetshowersandthunder": 25,
    "snow": 15,
    "snowandthunder": 25,
    "snowshowers": 21,
    "snowshowersandthunder": 25,
}


def get_condicode(code: str) -> Union[int, None]:
    """
    Map Met.no symbol codes to Meteostat condition codes

    Documentation: https://api.met.no/weatherapi/weathericon/2.0/documentation
    """
    return CONDICODES.get(str(code).split("_")[0], None)


def safe_get(data, keys, default=None, transform=lambda x: x):
    """
    Safely get a nested value from a dictionary, with an optional transformation.

    :param data: The dictionary to get the value from.
    :param keys: A list of keys to navigate through the dictionary.
    :param default: The default value to return if any key is missing.
    :param transform: A function to apply to the retrieved value if found.
    :return: The retrieved value or the default.
    """
    for key in keys:
        data = data.get(key)
        if data is None:
            return default
    return transform(data)


def map_data(record):
    """
    Map Met.no JSON data to Meteostat column names
    """
    details_instant = ["data", "instant", "details"]
    details_next_1_hour = ["data", "next_1_hours", "details"]

    return {
        "time": record["time"],
        Parameter.TEMP: safe_get(record, details_instant + ["air_temperature"]),
        Parameter.CLDC: safe_get(
            record,
            details_instant + ["cloud_area_fraction"],
            transform=percentage_to_okta,
        ),
        Parameter.RHUM: safe_get(record, details_instant + ["relative_humidity"]),
        Parameter.PRCP: safe_get(
            record, details_next_1_hour + ["precipitation_amount"]
        ),
        Parameter.WSPD: safe_get(
            record, details_instant + ["wind_speed"], transform=lambda x: x * 3.6
        ),
        Parameter.WPGT: safe_get(
            record,
            details_instant + ["wind_speed_of_gust"],
            transform=lambda x: x * 3.6,
        ),
        Parameter.WDIR: safe_get(
            record,
            details_instant + ["wind_from_direction"],
            transform=lambda x: int(round(x)),
        ),
        Parameter.PRES: safe_get(
            record, details_instant + ["air_pressure_at_sea_level"]
        ),
        Parameter.COCO: safe_get(
            record,
            ["data", "next_1_hours", "summary", "symbol_code"],
            transform=get_condicode,
        ),
    }


@cache_service.cache(TTL.HOUR, "pickle")
def get_df(latitude: float, longitude: float, elevation: int) -> Optional[pd.DataFrame]:
    endpoint = config.metno_forecast_endpoint
    user_agent = config.metno_user_agent

    if not endpoint:
        logger.warning("MET Norway forecast endpoint is not configured.")
        return None

    if not user_agent:
        logger.warning(
            "MET Norway requires a unique user agent as per their terms of service. Please use config to specify your user agent. For now, this provider is skipped."
        )
        return None

    file_url = endpoint.format(
        latitude=latitude,
        longitude=longitude,
        elevation=elevation,
    )

    headers = {"User-Agent": user_agent}

    try:
        response = network_service.get(file_url, headers=headers)

        # Raise an exception if the request was unsuccessful
        response.raise_for_status()

        # Parse the JSON content into a DataFrame
        data = response.json()

        # Create DataFrame
        df = pd.DataFrame(map(map_data, data["properties"]["timeseries"]))

        # Handle time column & set index
        df["time"] = pd.to_datetime(df["time"])
        df = df.set_index(["time"])

        # Remove the UTC timezone from the time index
        df.index = df.index.tz_localize(None)  # type: ignore[union-attr]

        # Shift prcp and coco columns by 1 (as they refer to the next hour)
        df["prcp"] = df["prcp"].shift(1)
        df["coco"] = df["coco"].shift(1)

        return df

    except HTTPError as error:
        logger.warning(
            f"Couldn't load weather forecast from met.no (status: {error.status})"
        )

    except Exception as error:
        logger.error(error, exc_info=True)


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    return get_df(
        req.station.latitude,
        req.station.longitude,
        req.station.elevation,
    )



"""
DWD Global CLIMAT Data
"""

from datetime import datetime
from ftplib import FTP
from io import BytesIO
from typing import List, Optional

import pandas as pd

from meteostat.core.logger import logger
from meteostat.api.config import config
from meteostat.enumerations import TTL, Parameter
from meteostat.typing import ProviderRequest, Station
from meteostat.core.cache import cache_service
from meteostat.providers.dwd.shared import get_ftp_connection

# Constants
BASE_DIR = "/climate_environment/CDC/observations_global/CLIMAT/monthly/qc/"

# Monthly column stubnames mapping template
MONTHS_MAP = {
    "jan": 1,
    "feb": 2,
    "mrz": 3,
    "apr": 4,
    "mai": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "okt": 10,
    "nov": 11,
    "dez": 12,
}

# Parameter directory configurations
PARAMETERS = [
    ("precipitation_total", Parameter.PRCP),
    ("air_temperature_mean", Parameter.TEMP),
    ("air_temperature_mean_of_daily_max", Parameter.TMAX),
    ("air_temperature_mean_of_daily_min", Parameter.TMIN),
    ("air_temperature_absolute_max", Parameter.TXMX),
    ("air_temperature_absolute_min", Parameter.TXMN),
    ("mean_sea_level_pressure", Parameter.PRES),
    ("sunshine_duration", Parameter.TSUN),
]

# Precomputed parameter configs with stubnames
PARAMETER_CONFIGS = {
    param: {
        "dir": dir_name,
        "stubnames": {
            "jahr": "year",
            **{
                month: f"{param}{i + 1}"
                for i, (month, _) in enumerate(MONTHS_MAP.items())
            },
        },
    }
    for dir_name, param in PARAMETERS
}


def find_file(ftp: FTP, mode: str, directory: str, search_term: str) -> Optional[str]:
    """
    Find a file in the FTP directory matching a pattern.
    """
    try:
        ftp.cwd(f"{BASE_DIR}{directory}/{mode}")
        matches = [f for f in ftp.nlst() if search_term in f]
        return sorted(matches)[-1] if matches else None
    except Exception:
        logger.debug("Error while searching for file", exc_info=True)
        return None


@cache_service.cache(TTL.WEEK, "pickle")
def get_df(parameter: str, mode: str, station_code: str) -> Optional[pd.DataFrame]:
    """
    Download and parse a CLIMAT dataset from DWD FTP.
    """
    param_config = PARAMETER_CONFIGS.get(parameter)
    if not param_config:
        logger.debug(f"Unknown parameter '{parameter}'")
        return None

    ftp = get_ftp_connection()
    search_term = station_code if mode == "recent" else f"{station_code}_"
    remote_file = find_file(ftp, mode, param_config["dir"], search_term)

    if not remote_file:
        logger.debug(
            f"No file found for parameter '{parameter}', mode '{mode}', station '{station_code}'"
        )
        return None

    buffer = BytesIO()
    ftp.retrbinary(f"RETR {remote_file}", buffer.write)
    ftp.close()

    buffer.seek(0)
    df = pd.read_csv(buffer, sep=";").rename(columns=lambda col: col.strip().lower())
    df = df.rename(columns=param_config["stubnames"])

    # Convert wide to long format
    df = pd.wide_to_long(
        df, stubnames=parameter, i="year", j="month", sep="", suffix="\\d+"
    ).reset_index()

    if parameter == Parameter.TSUN:
        df[Parameter.TSUN] *= 60  # convert hours to minutes

    # Create datetime index
    df["time"] = pd.to_datetime(
        df["year"].astype(str) + "-" + df["month"].astype(str).str.zfill(2) + "-01"
    )

    return df.drop(columns=["year", "month"]).set_index("time")


def get_parameter(
    parameter: str, modes: List[str], station: Station
) -> Optional[pd.DataFrame]:
    """
    Fetch and merge data for a parameter over multiple modes (e.g., recent, historical).
    """
    try:
        station_code = station.identifiers.get("wmo")
        if not station_code:
            return None

        datasets = [get_df(parameter, mode, station_code) for mode in modes]
        datasets = [df for df in datasets if df is not None]

        if not datasets:
            return None

        return pd.concat(datasets).loc[lambda df: ~df.index.duplicated(keep="first")]
    except Exception as e:
        logger.warning(
            f"Failed to fetch data for parameter '{parameter}': {e}", exc_info=True
        )
        return None


def fetch(req: ProviderRequest) -> pd.DataFrame:
    """
    Entry point to fetch all requested parameters for a station query.
    """
    station_code = req.station.identifiers.get("wmo")
    if not station_code:
        return pd.DataFrame()

    modes = ["historical"]
    if (datetime.now() - req.end).days < 5 * 365:
        modes.append("recent")

    data_frames = [
        get_parameter(param.value, config.dwd_climat_modes or modes, req.station)
        for param in req.parameters
        if param in PARAMETER_CONFIGS
    ]

    return pd.concat([df for df in data_frames if df is not None], axis=1)



"""
Guard functions for Meteostat.

The code is licensed under the MIT license.
"""

from datetime import datetime
from meteostat.api.config import config
from meteostat.core.logger import logger
from meteostat.enumerations import Granularity
from meteostat.typing import Request


def request_size_guard(req: Request) -> None:
    """
    Guard to block large requests
    """
    if not config.block_large_requests:
        logger.debug("Large request blocking is disabled.")
        return

    if isinstance(req.station, list) and len(req.station) > 10:
        raise ValueError(
            "Requests with more than 10 stations are blocked by default. "
            "To enable large requests, set `config.block_large_requests = False`."
        )

    if req.granularity not in [Granularity.HOURLY, Granularity.DAILY]:
        return

    if req.start is None:
        raise ValueError(
            "Hourly and daily requests without a start date are blocked by default. "
            "To enable large requests, set `config.block_large_requests = False`."
        )

    time_diff_years = abs((req.end or datetime.now()).year - req.start.year)

    logger.debug(f"Request time range: {time_diff_years} years.")

    if req.granularity is Granularity.HOURLY and time_diff_years > 3:
        raise ValueError(
            "Hourly requests longer than 3 years are blocked by default. "
            "To enable large requests, set `config.block_large_requests = False`."
        )

    if req.granularity is Granularity.DAILY and time_diff_years > 30:
        raise ValueError(
            "Daily requests longer than 30 years are blocked by default. "
            "To enable large requests, set `config.block_large_requests = False`."
        )
